(**************************************************************)
(*   The Strong Completeness of Partial Monoidal BI           *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.

Require Import list_fall_exst.
Require Import predicate.

Require Import labels.
Require Import formula.
Require Import words.
Require Import constraints.
Require Import pmo.
Require Import css.

Section BI_expansion.

  Variable p : (formula*formula*label*label*word label*word label*word label)%type.
  
  Let f := match p with (f,_,_,_,_,_,_) => f end.
  Let g := match p with (_,g,_,_,_,_,_) => g end.
  Let a := match p with (_,_,a,_,_,_,_) => a end.
  Let b := match p with (_,_,_,b,_,_,_) => b end.
  Let m := match p with (_,_,_,_,m,_,_) => m end.
  Let x := match p with (_,_,_,_,_,x,_) => x end.
  Let y := match p with (_,_,_,_,_,_,y) => y end.
  
  Variable css : CSS_type.
  
  Let X := match css with (X,_) => X end.
  Let C := match css with (_,C) => C end.

  Notation ec := (Emptyset cst).
  Notation es := (Emptyset stm).
  Notation eps := (nil : word label).
  
  Definition bi_exp_T_and_cond := X (T f && g # m).
  Definition bi_exp_T_and_inst := (sg (T f # m) cup sg (T g # m), ec)::nil.
  
  Definition bi_exp_F_and_cond := X (F f && g # m).
  Definition bi_exp_F_and_inst := (sg (F f # m), ec)::(sg (F g # m), ec)::nil.

  Definition bi_exp_F_or_cond := X (F f %% g # m).
  Definition bi_exp_F_or_inst := (sg (F f # m) cup sg (F g # m), ec)::nil.
  
  Definition bi_exp_T_or_cond := X (T f %% g # m).
  Definition bi_exp_T_or_inst := (sg (T f # m), ec)::(sg (T g # m), ec)::nil.

  Definition bi_exp_F_imp_cond := X (F f #> g # m) /\ ~ alphabet C b.
  Definition bi_exp_F_imp_inst := (sg (T f # l b) cup sg (F g # l b), sg (m ~~ l b))::nil.
  
  Definition bi_exp_T_imp_cond := X (T f #> g # m) /\ pmo C (m ~~ x).
  Definition bi_exp_T_imp_inst := (sg (F f # x), ec)::(sg (T g # x), ec)::nil.
  
  Definition bi_exp_T_unit_cond := X (T I # m).
  Definition bi_exp_T_unit_inst := (es, sg (eps~~m))::nil.
  
  Definition bi_exp_T_mult_cond := X (T (f * g) # m) /\ a <> b /\ ~ alphabet C a /\ ~ alphabet C b.
  Definition bi_exp_T_mult_inst := (sg (T f # l a) cup sg (T g # l b), sg (l a o l b ~~ m))::nil.
  
  Definition bi_exp_F_mult_cond := X (F (f * g) # m) /\ pmo C (x o y ~~ m).
  Definition bi_exp_F_mult_inst := (sg (F f # x), ec)::(sg (F g # y), ec)::nil.
  
  Definition bi_exp_T_wand_cond := X (T (f -* g) # m) /\ pmo C (x o m ~~ y).
  Definition bi_exp_T_wand_inst := (sg (F f # x), ec)::(sg (T g # y), ec)::nil.
  
  Definition bi_exp_F_wand_cond := X (F (f -* g) # m) /\ a <> b /\ ~ alphabet C a /\ ~ alphabet C b.
  Definition bi_exp_F_wand_inst := (sg (T f # l a) cup sg (F g # l b), sg (l a o m ~~ l b))::nil.
  
End BI_expansion.

Record bi_tab_rule : Type := MkBI_rule {
  bi_param : Type;
  bi_cond  : bi_param -> CSS_type -> Prop;
  bi_inst  : bi_param -> list CSS_type
}.

Implicit Arguments MkBI_rule [ bi_param ].

Definition bi_exp_T_and  := MkBI_rule bi_exp_T_and_cond  bi_exp_T_and_inst.
Definition bi_exp_F_and  := MkBI_rule bi_exp_F_and_cond  bi_exp_F_and_inst.
Definition bi_exp_T_or   := MkBI_rule bi_exp_T_or_cond   bi_exp_T_or_inst.
Definition bi_exp_F_or   := MkBI_rule bi_exp_F_or_cond   bi_exp_F_or_inst.
Definition bi_exp_T_imp  := MkBI_rule bi_exp_T_imp_cond  bi_exp_T_imp_inst.
Definition bi_exp_F_imp  := MkBI_rule bi_exp_F_imp_cond  bi_exp_F_imp_inst.
Definition bi_exp_T_unit := MkBI_rule bi_exp_T_unit_cond bi_exp_T_unit_inst.
Definition bi_exp_T_mult := MkBI_rule bi_exp_T_mult_cond bi_exp_T_mult_inst.
Definition bi_exp_F_mult := MkBI_rule bi_exp_F_mult_cond bi_exp_F_mult_inst.
Definition bi_exp_T_wand := MkBI_rule bi_exp_T_wand_cond bi_exp_T_wand_inst.
Definition bi_exp_F_wand := MkBI_rule bi_exp_F_wand_cond bi_exp_F_wand_inst.

Definition bi_exp_rules := 
    bi_exp_T_and::bi_exp_F_and::bi_exp_T_or::bi_exp_F_or::bi_exp_T_imp::bi_exp_F_imp::
    bi_exp_T_unit::bi_exp_T_mult::bi_exp_F_mult::bi_exp_T_wand::bi_exp_F_wand::nil.

Definition expand_1 (css1 css2 : CSS_type) := match (css1,css2) with ((X,C),(X1,C1)) => (X cup X1, C cup C1) end.
Definition expand CSS lCSS := map (expand_1 CSS) lCSS.

Notation bi_fireable := bi_cond.

Inductive bi_expansion css : list CSS_type -> Prop := 
  | In_bi_exp : forall r p,      In r bi_exp_rules 
                               -> bi_fireable r p css
                               -> bi_expansion css (expand css (bi_inst r p)).

Fixpoint tab_constraints (lcss : list CSS_type) :=
  match lcss with
    | nil         => empty
    | (X,C)::lcss => C cup (tab_constraints lcss)
  end.

Section BI_exp_basic.

  Variable X : stm -> Prop.
  Variable C : cst -> Prop.

  Ltac solve r1 p1 := 
    intros;
    apply In_bi_exp with (css := (X,C)) (r := r1) (p := p1); 
    [ unfold bi_exp_rules; In_auto
    | simpl; auto ]; red; auto.

  Proposition bi_exp_Tand A B m : X (T A && B # m) -> bi_expansion (X,C) ((X cup (sg (T A # m) cup sg (T B # m)), C cup empty)::nil).
  Proof. solve bi_exp_T_and (A,B,l0,l0,m,m,m). Qed.
  
  Proposition bi_exp_Fand A B m : X (F A && B # m) -> bi_expansion (X,C) ((X cup sg (F A # m), C cup empty)::
                                                                          (X cup sg (F B # m), C cup empty)::nil).
  Proof. solve bi_exp_F_and (A,B,l0,l0,m,m,m). Qed.

  Proposition bi_exp_Tor A B m : X (T A %% B # m) -> bi_expansion (X,C) ((X cup sg (T A # m), C cup empty)::
                                                                         (X cup sg (T B # m), C cup empty)::nil).
  Proof. solve bi_exp_T_or (A,B,l0,l0,m,m,m). Qed.  

  Proposition bi_exp_For A B m : X (F A %% B # m) -> bi_expansion (X,C) ((X cup (sg (F A # m) cup sg (F B # m)), C cup empty)::nil).
  Proof. solve bi_exp_F_or (A,B,l0,l0,m,m,m). Qed.              

  Proposition bi_exp_Timp A B x m : X (T A #> B # m) -> pmo C (m ~~ x)
                                            -> bi_expansion (X,C) ((X cup sg (F A # x), C cup empty)::
                                                                   (X cup sg (T B # x), C cup empty)::nil).
  Proof. solve bi_exp_T_imp (A,B,l0,l0,m,x,x). Qed.

  Proposition bi_exp_Fimp A B b m : X (F A #> B # m) -> ~ alphabet C b 
                                             -> bi_expansion (X,C) ((X cup (sg (T A # l b) cup sg (F B # l b)), C cup sg (m ~~ l b))::nil).
  Proof. solve bi_exp_F_imp (A,B,b,b,m,m,m). Qed.

  Proposition bi_exp_TI m : X (T I # m) -> bi_expansion (X,C) ((X cup empty, C cup sg (eps ~~ m))::nil).
  Proof. solve bi_exp_T_unit (I,I,l0,l0,m,m,m). Qed.

  Proposition bi_exp_Tmult A B a b m : X (T A * B # m) -> a<> b -> ~ alphabet C a -> ~ alphabet C b 
                                             -> bi_expansion (X,C) ((X cup (sg (T A # l a) cup sg (T B # l b)), C cup sg (l a o l b ~~ m))::nil).
  Proof. solve bi_exp_T_mult (A,B,a,b,m,m,m). Qed.

  Proposition bi_exp_Fmult A B x y m : X (F A * B # m) -> pmo C (x o y ~~ m)
                                             -> bi_expansion (X,C) ((X cup sg (F A # x), C cup empty)::
                                                                    (X cup sg (F B # y), C cup empty)::nil).
  Proof. solve bi_exp_F_mult (A,B,l0,l0,m,x,y). Qed.

  Proposition bi_exp_Twand A B x y m : X (T A -* B # m) -> pmo C (x o m ~~ y)
                                             -> bi_expansion (X,C) ((X cup sg (F A # x), C cup empty)::
                                                                    (X cup sg (T B # y), C cup empty)::nil).
  Proof. solve bi_exp_T_wand (A,B,l0,l0,m,x,y). Qed.

  Proposition bi_exp_Fwand A B a b m : X (F A -* B # m) -> a<> b -> ~ alphabet C a -> ~ alphabet C b 
                                             -> bi_expansion (X,C) ((X cup (sg (T A # l a) cup sg (F B # l b)), C cup sg (l a o m ~~ l b))::nil).
  Proof. solve bi_exp_F_wand (A,B,a,b,m,m,m). Qed.

End BI_exp_basic.
