(**************************************************************)
(*   The Strong Completeness of Partial Monoidal BI           *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.
Require Import Classical.

(* The Soundness and Completeness of Kripke semantics cannot be proved without classical logic *)

Require Import classical_help.
Require Import list_fall_exst.
  
Require Import predicate.
Require Import closure.
Require Import carrier.
Require Import finite.

Require Import labels.
Require Import formula.
Require Import words.
Require Import substitutions.
Require Import constraints.
Require Import css.
Require Import pmo.

Require Import kripke.

Require Import branch.
Require Import tableaux.

Require Import bi_expansion.
Require Import bi_exp_props.
Require Import bi_tableaux.
Require Import bi_closed_branch.

Definition realizable (css : CSS_type) := CSS css 
                                       /\ exists L, 
                                          exists meq : constraint L -> Prop, 
                                          exists fV, 
                                          exists rho, bi_frame meq fV /\ mdl_CSS meq fV rho css.

Proposition realizable_id : forall css1 css2, CSS_id css1 css2 -> realizable css1 -> realizable css2.
Proof.
  intros (X1,C1) (X2,C2) [ HX HC ] [ H (L & meq & fV & rho & HM1 & HM2 & HM3 & HM4) ]; split.
  apply (CSS_id_id (X1,C1)); auto.
  exists L; exists meq; exists fV; exists rho.
  split; auto.
  split.
  generalize H; apply CSS_id_id; auto.
  split.
  apply inc_transitive with (2 := HM3), HX; auto.
  apply inc_transitive with (2 := HM4), HC; auto.
Qed.

Hint Resolve realizable_id.

Proposition realizable_not_closed : forall css, realizable css -> bi_closed_branch css -> False.
Proof.
  intros css [ Hcss ( L & meq & fV & rho & HM1 & HM2 ) ] H.
  apply bin_inter_sg in H; destruct H as [ _ H ].    
  destruct H as [ X C A m n H1 H2 H3 | X C m H1 H2 | X C m H ];
  destruct HM2 as ( _ & HX & HC ).

  apply HX in H1.
  apply HX in H2; auto. 
  apply meq_mdl_cst with (ivar := fV) (2 := HC) in H3; auto.
  simpl in H1, H2, H3.
  apply H2, forces_monotonic with (1 := HM1) (3 := H1); auto.
  
  apply HX in H1.
  apply meq_mdl_cst with (ivar := fV) (2 := HC) in H2; auto.

  apply HX in H; simpl in H; auto.
Qed.

(* Hint Resolve mdl_interp_mono. *)

Section Realizability.

  Let bbi_subst L (meq : constraint L -> Prop) fV rho C m n : bi_frame meq fV -> pmo C (m~~n) -> C inc mdl_cst meq rho -> meq (subst_w rho m ~~ subst_w rho n).
  Proof.
    intros [ ] H0 H1; simpl in H1 |- *. 
    apply pmo_min in H1; auto.
    apply H1 in H0; auto.
    apply mdl_cst_pmo with (ivar := fV).
    split; auto.
  Qed.

  Proposition realizable_stable : expansion_stable CSS_id bi_expansion realizable.
  Proof.
    intros (X,C) lb Hlb [ HXC (L & meq & fV & rho & H1 & H2 & HX & HC) ].

    assert (carrier lb inc CSS) as HCSS. apply bi_expansion_CSS with (2 := Hlb); auto.

    destruct Hlb as [ r p Hr Hp ];
    unfold bi_exp_rules in Hr; In_auto; symmetry in Hr; 
    subst; destruct p as ((((((A,B),a),b),m),x),y);
    simpl in Hp, HCSS; try red in Hp; simpl; unfold expand_1 in HCSS |- *;
    try ( match goal with [ |- exists _, exists _, ( ?x = _ \/ False) /\ _ /\ _ ] => exists x; exists x; assert (CSS x) as HCSS'; [ apply HCSS; auto | idtac ] end;
          split; auto; split; auto; split; auto).
  
    (* (M,rho) |= T A & B # m *)
    
    exists L; exists meq; exists fV; exists rho.
    split; auto.
    apply HX in Hp; simpl in Hp.
    split; pred_solve; tauto.
    
    (* (M,rho) |= F A & B # m *)
    
    apply HX in Hp; simpl in Hp.
    classic_solve;
    destruct Hp as [ Hp | Hp ].
    
    do 2 exists (X cup sg (F A # m),C cup empty).
    split; auto. split; auto.
    assert (CSS (X cup sg (F A # m),C cup empty)) as HCSS'. apply HCSS; auto.
    split; auto. 
    exists L; exists meq; exists fV; exists rho. 
    repeat (split; auto).
    
    do 2 exists (X cup sg (F B # m),C cup empty).
    split; auto; split; auto.
    assert (CSS (X cup sg (F B # m),C cup empty)) as HCSS'. apply HCSS; auto.
    split; auto. 
    exists L; exists meq; exists fV; exists rho. 
    repeat (split; auto).

    (* (M,rho) |= T A %% B # m *)
    
    apply HX in Hp; simpl in Hp.
    classic_solve;
    destruct Hp as [ Hp | Hp ].
    
    do 2 exists (X cup sg (T A # m),C cup empty).
    split; auto. split; auto.
    assert (CSS (X cup sg (T A # m),C cup empty)) as HCSS'. apply HCSS; auto.
    split; auto. 
    exists L; exists meq; exists fV; exists rho. 
    repeat (split; auto).
    
    do 2 exists (X cup sg (T B # m),C cup empty).
    split; auto; split; auto.
    assert (CSS (X cup sg (T B # m),C cup empty)) as HCSS'. apply HCSS; auto.
    split; auto. 
    exists L; exists meq; exists fV; exists rho. 
    repeat (split; auto).

    (* (M,rho) |= F A %% B # m *)
    
    exists L; exists meq; exists fV; exists rho.
    split; auto.
    apply HX in Hp; simpl in Hp.
    classic_solve.
    split; pred_solve; tauto.

    (* (M,rho) |= T A #> B # m *)
    
    destruct Hp as ( Hp1 & Hp2 ).
    apply HX in Hp1; simpl in Hp1.
    specialize Hp1 with (x := subst_w rho x).
    classic_solve;
    destruct Hp1 as [ Hp1 | Hp1 ].

    exfalso; apply Hp1.
    apply meq_mdl_cst with (1 := H1) in HC.
    apply HC in Hp2.
    apply Hp2.

    classic_solve.
    destruct Hp1 as [ Hp1 | Hp1 ].
    
    do 2 exists (X cup sg (F A # x),C cup empty).
    split; auto. split; auto.
    assert (CSS (X cup sg (F A # x),C cup empty)) as HCSS'. apply HCSS; auto.
    split; auto. 
    exists L; exists meq; exists fV; exists rho. 
    repeat (split; auto).
    
    do 2 exists (X cup sg (T B # x),C cup empty).
    split; auto; split; auto.
    assert (CSS (X cup sg (T B # x),C cup empty)) as HCSS'. apply HCSS; auto.
    split; auto. 
    exists L; exists meq; exists fV; exists rho. 
    repeat (split; auto).

    (* (M,rho) |= F A #> B # m *)
    
    destruct Hp as [ Hp1 Hp2 ].
    generalize (HX _ Hp1); intros Hp1'; simpl in Hp1'.
    classic_solve.
    destruct Hp1' as ( x' & H1' & H2' & H3' ).
    
    set (rho' := change_two rho b x' b x').
    assert (rho' b = x') as Hby. unfold rho'; apply change_two_prop_a; auto.
    assert (mdl_CSS meq fV rho' (X,C)) as HXC'.
      apply meq_mdl_eq with (rho := rho).
      intros c Hc; unfold rho'.
      symmetry; apply change_two_l_prop_ab; intros Hc'; rewrite Hc' in Hc; contradict Hc; auto.
      split; auto.
      
    destruct HXC' as [ HX' HC' ].

    exists L; exists meq; exists fV; exists rho'.
    repeat (split; pred_solve);
    simpl; rewrite Hby, eps_neutral_r; auto.
    apply H2, pmo_inc_alphabet in Hp1.
    unfold rho'; rewrite change_two_prop_carrier; auto.

    (* (M,rho) |= T I # m *)
    
    exists L; exists meq; exists fV; exists rho.
    split; auto.
    repeat (split; pred_solve).
    apply HX in Hp; simpl in Hp; tauto.
    
    (* (M,rho) |= T A * B # m *)
    
    destruct Hp as ( Hp1 & Hp2 & Hp3 & Hp4).
    generalize (HXC _ _ _ Hp1); intros Hp1'.
    apply HX in Hp1; simpl in Hp1.
    destruct Hp1 as ( x' & y' & H1' & H2' & H3' ).
    
    set (rho' := change_two rho a x' b y').
    assert (rho' a = x') as Hax. unfold rho'; apply change_two_prop_a.
    assert (rho' b = y') as Hby. unfold rho'; apply change_two_prop_b; auto.
    assert (mdl_CSS meq fV rho' (X,C)) as HXC'.
      apply meq_mdl_eq with (rho := rho).
      intros c Hc; unfold rho'.
      symmetry; apply change_two_l_prop_ab; intros Hc'; rewrite Hc' in Hc; contradict Hc; auto.
      split; auto.

    destruct HXC' as [ HX' HC' ].

    exists L; exists meq; exists fV; exists rho'.
    repeat (split; pred_solve); simpl; 
    try rewrite Hax; try rewrite Hby; try rewrite eps_neutral_r; auto.
    unfold rho'; rewrite change_two_prop_carrier; auto;
    [ contradict Hp3 | contradict Hp4 ]; 
    apply pmo_inc_alphabet with (m := m); auto.

    (* (M,rho) |= F A * B # m *)
    
    destruct Hp as ( Hp1 & Hp2 ).
    generalize (HXC _ _ _ Hp1); intros Hp1'.
    apply HX in Hp1; simpl in Hp1.
    generalize (not_ex_ex _ _ _ Hp1 (subst_w rho x) (subst_w rho y)); clear Hp1; intros Hp1.

    classic_solve.
    destruct Hp1 as [ Hp1 | Hp1 ].
    
    apply meq_mdl_cst with (1 := H1) (2 := HC) in Hp2. 
    simpl in Hp2. rewrite subst_o in Hp2. 
    contradiction (Hp1 Hp2).

    classic_solve.
    destruct Hp1 as [ Hp1 | Hp1 ].
    
    do 2 exists (X cup sg (F A # x), C cup empty).
    do 3 (split; auto).
    exists L; exists meq; exists fV; exists rho.
    repeat (split; auto).
    
    do 2 exists (X cup sg (F B # y), C cup empty).
    do 3 (split; auto).
    exists L; exists meq; exists fV; exists rho.
    repeat (split; auto).

    (* (M,rho) |= T A -* B # m *)
   
    destruct Hp as ( Hp1 & Hp2 ).
    generalize (HXC _ _ _ Hp1); intros Hp1'.
    apply HX in Hp1; simpl in Hp1.
    generalize (Hp1 (subst_w rho x) (subst_w rho y)); clear Hp1; intros Hp1.
    apply meq_mdl_cst with (1 := H1) (2 := HC) in Hp2.
    simpl in Hp2.
    rewrite subst_o in Hp2. 
    generalize (Hp1 Hp2); clear Hp1; intros Hp1.

    classic_solve.
    destruct Hp1 as [ Hp1 | Hp1 ].

    do 2 exists (X cup sg (F A # x), C cup empty).
    do 3 (split; auto).
    exists L; exists meq; exists fV; exists rho.
    repeat (split; auto).

    do 2 exists (X cup sg (T B # y), C cup empty).
    do 3 (split; auto).
    exists L; exists meq; exists fV; exists rho.
    repeat (split; auto).

    (* (M,rho) |= F A -* B # m *)

    destruct Hp as ( Hp1 & Hp2 & Hp3 & Hp4).
    generalize (HXC _ _ _ Hp1); intros Hp1'.
    apply HX in Hp1; simpl in Hp1.
    
    classic_solve.
    destruct Hp1 as ( x' & y' & H1' & H2' & H3' ).
    
    set (rho' := change_two rho a x' b y').
    assert (rho' a = x') as Hax. unfold rho'; apply change_two_prop_a.
    assert (rho' b = y') as Hby. unfold rho'; apply change_two_prop_b; auto.
    assert (mdl_CSS meq fV rho' (X,C)) as HXC'.
      apply meq_mdl_eq with (rho := rho).
      intros c Hc; unfold rho'.
      symmetry; apply change_two_l_prop_ab; intros Hc'; rewrite Hc' in Hc; contradict Hc; auto.
      split; auto.

    destruct HXC' as [ HX' HC' ].

    exists L; exists meq; exists fV; exists rho'.
    repeat (split; pred_solve); simpl; 
    try rewrite Hax; try rewrite Hby; try rewrite eps_neutral_r; auto.
    unfold rho'; rewrite change_two_prop_carrier; auto;
    [ contradict Hp3 | contradict Hp4 ]; 
    apply pmo_inc_alphabet with (m := m); auto.
  Qed.

End Realizability.
