(**************************************************************)
(*   The Strong Completeness of Partial Monoidal BI           *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.

Require Import predicate.
Require Import carrier.

Section list_lift.

  Variable X Y : Type.
  Variable P : X -> Prop.
  Variable Q : Y -> Prop.

  Variable f : X -> Y.
  Hypothesis Hf : forall x, P x -> Q (f x).

  Proposition list_lift l : carrier l inc P -> carrier (map f l) inc Q.
  Proof.
    induction l; simpl; auto.
    intros H; apply carrier_cons_1 in H; destruct H.
    apply carrier_cons_2; auto.
  Qed.

End list_lift.

Section listbranch_eq.

  Variable X : Type.
  Variable eq : X -> X -> Prop.

  Fixpoint listbranch_eq (l1 l2 : list X) { struct l1 } := 
    match (l1,l2) with
      | (nil,nil) => True
      | (b1::l1,b2::l2) => eq b1 b2 /\ listbranch_eq l1 l2
      | _ => False
    end.

  Infix "leq" := listbranch_eq (at level 60).

  Proposition listbranch_eq_app : forall l1 l1' l2 l2', l1 leq l1' -> l2 leq l2' -> (l1++l2) leq (l1'++l2').
  Proof.
    induction l1 as [ | b1 l1 ]; induction l1' as [ | b1' l1' ]; intros l2 l2' H1 H2; auto; try contradiction H1. 
    simpl in H1 |- *. 
    split. tauto.
    apply (IHl1 l1' l2 l2'); tauto.
  Qed.

  Hint Resolve listbranch_eq_app.
  
  Proposition listbranch_eq_app_destruct : forall l1 l2 lb, (l1++l2) leq lb -> exists l1', 
                                                     exists l2', l1 leq l1' 
                                                              /\ l2 leq l2' 
                                                              /\ lb = l1'++l2'.
  Proof.
    induction l1 as [ | b1 l1 ]; intros l2 lb H.
    simpl in H. exists nil. exists lb. simpl; auto. 
    induction lb as [ | b lb ]. contradiction H.
    simpl in H. destruct H as [ HCSS H ].
    destruct (IHl1 _ _ H) as [ l1' [ l2' [ Hl1 [ Hl2 Hlb ] ] ] ].
    exists (b::l1'). exists l2'.
    simpl; rewrite Hlb; auto.
  Qed.
  
  Proposition listbranch_eq_split : forall l1 b l2 lb, (l1++b::l2) leq lb 
                                                      -> exists l1', 
                                                         exists b', 
                                                         exists l2', l1 leq l1' 
                                                                  /\ eq b b' 
                                                                  /\ l2 leq l2' 
                                                                  /\ lb = l1'++b'::l2'.
  Proof.
    intros l1 b l2 lb H.
    cutrewrite (l1++b::l2 = l1++(b::nil)++l2) in H; auto.
    destruct (listbranch_eq_app_destruct _ _ _ H) as [ l1' [ l' [ H0 [ H1 H2 ] ] ] ].
    destruct (listbranch_eq_app_destruct _ _ _ H1) as [ l'' [ l2' [ H3 [ H4 H5 ] ] ] ].
    induction l'' as [ | b' l'' ]. contradiction H3.
    induction l''.
    clear IHl''. exists l1'. exists b'. exists l2'. simpl in H3.
    split. trivial. split. tauto. split; trivial.
    rewrite H2, H5; auto.
    simpl in H3; tauto.
  Qed.

  Section listbranc_eq_map.
  
    Variable f : X -> X.
    Hypothesis Hf : forall x y, eq x y -> eq (f x) (f y).

    Proposition listbranch_eq_map l1 : forall l2, listbranch_eq l1 l2 -> listbranch_eq (map f l1) (map f l2).
    Proof. induction l1 as [ | x l1 ]; intros [ | y l2 ]; auto; intros [ ]; simpl; split; auto. Qed.

  End listbranc_eq_map.

  Variable P : X -> Prop.
  Hypothesis HP : forall x y, eq x y -> P x -> P y.
  
  Proposition listbranch_eq_lift l : forall m, l leq m -> carrier l inc P -> carrier m inc P.
  Proof.
    induction l as [ | x l IHl ]; destruct m as [ | y m ]; auto.
    intros H; contradiction H.
    simpl.
    intros [ H1 H2 ] H3.
    apply carrier_cons_2.
    apply (HP _ _ H1), H3; auto.
    apply IHl; auto.
    apply inc_transitive with (2 := H3).
    intro; apply carrier_cons_tail.
  Qed.

End listbranch_eq.

Implicit Arguments listbranch_eq [ X ].

Hint Resolve listbranch_eq_app.