(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.

Require Import predicate.

Definition carrier {X} l (x : X) := In x l.

Section Carrier.

  Variable X : Type.
  Hypothesis eqX_dec : forall x y : X, { x = y } + { x <> y }.
  
  Proposition carrier_dec l (x : X) : { carrier l x } + { ~ carrier l x }.
  Proof. apply (in_dec eqX_dec). Qed.

  Definition carrier_bool l x := match carrier_dec l x with left _ => true | right _ => false end.

  Proposition carrier_dec_bool l : carrier l ~ (fun x => carrier_bool l x = true).
  Proof.
    unfold carrier_bool.
    split; intros x H;
    destruct (carrier_dec l x) as [ | H' ]; auto.
    discriminate H.
  Qed.

  Hint Immediate carrier_dec_bool.

  Proposition carrier_nil_1 (x : X) : (carrier nil) x -> False.
  Proof. intros []. Qed.

  Proposition carrier_nil_2 C : carrier (X := X) nil inc C.
  Proof. intros ? []. Qed.

  Hint Resolve carrier_nil_1.
  Hint Immediate carrier_nil_2.

  Proposition carrier_nil : carrier (X:=X) nil ~ empty.
  Proof. split; auto. Qed.

  Hint Immediate carrier_nil.

  Proposition carrier_cons_hd l (x : X) : carrier (x::l) x.
  Proof. left; auto. Qed.

  Proposition carrier_cons_tail l (x y: X) : carrier l x -> carrier (y::l) x.
  Proof. right; auto. Qed.
  
  Hint Immediate carrier_cons_hd.
  Hint Resolve carrier_cons_tail.

  Proposition carrier_cons l (x : X) : carrier (x::l) ~ sg x cup carrier l.
  Proof.
    split. 
    intros y Hy; apply in_inv in Hy.
    destruct Hy; pred_solve.
    pred_solve.
    intro; unfold carrier, In; auto.
  Qed.

  Proposition carrier_cons_1 C l (x : X) : carrier (x::l) inc C -> C x /\ carrier l inc C.
  Proof. 
    intros H.
    apply inc_transitive with (1 := proj2 (carrier_cons _ _)) in H.
    pred_solve.
  Qed.

  Proposition carrier_cons_2 (C : X -> Prop) l x : C x -> carrier l inc C -> carrier (x::l) inc C.
  Proof.
    intros.
    apply inc_transitive with (1 := proj1 (carrier_cons _ _)).
    pred_solve.
  Qed.

  Hint Resolve carrier_cons_2.

  Proposition carrier_nil_3 (l : list X) : carrier l inc empty -> l = nil.
  Proof.
    induction l; auto. 
    intros H; apply carrier_cons_1 in H; pred_solve.
  Qed.
  
  Proposition carrier_app (l l' : list X) : carrier (l++l') ~ carrier l cup carrier l'.
  Proof.
    split.
    intros x Hx. 
    unfold carrier in Hx; apply in_app_or in Hx.
    destruct Hx; pred_solve.
    intros x Hx; apply in_or_app; pred_solve.
  Qed.

  Hint Resolve carrier_app.

  Proposition carrier_app_1 C (l l' : list X) : carrier (l++l') inc C -> (carrier l inc C) /\ (carrier l' inc C).
  Proof. intros H; apply bin_union_inc_r, inc_transitive with (2 := H), carrier_app. Qed.
 
  Proposition carrier_app_2 C (l l' : list X) : carrier l inc C -> carrier l' inc C -> carrier (l++l') inc C.
  Proof. intros; apply inc_transitive with (1 := proj1 (carrier_app _ _)); auto. Qed.

  Hint Resolve carrier_app_2.

  Proposition carrier_app_lft (l l' : list X) : carrier l inc carrier (l++l').
  Proof. apply inc_transitive with (2 := proj2 (carrier_app _ _)); auto. Qed.

  Proposition carrier_app_rt (l l' : list X) : carrier l' inc carrier (l++l').
  Proof. apply inc_transitive with (2 := proj2 (carrier_app _ _)); auto. Qed.

  Hint Immediate carrier_app_lft carrier_app_rt.

  Proposition carrier_app_disj l l' (x : X) : (carrier (l++l')) x -> (carrier l) x \/ (carrier l') x.
  Proof. intros; apply bin_union_sg, carrier_app; auto. Qed.

  Proposition carrier_cap_1 (P Q : X -> Prop) ll : carrier ll inc P -> carrier ll inc Q -> carrier ll inc (P cap Q) .
  Proof. intros; pred_solve. Qed.

  Proposition carrier_map (f : X -> X) (P : _ -> Prop) ll : (forall x, P x -> P (f x)) -> carrier ll inc P -> carrier (map f ll) inc P.
  Proof. 
    intros HP Hll. 
    induction ll; simpl; auto.
    apply carrier_cons_1 in Hll.
    destruct Hll as [ ? Hll ].
    apply IHll in Hll.
    apply carrier_cons_2; auto.
  Qed.

  Proposition carrier_filter (l : list X) f : carrier (filter f l) ~ carrier l cap (fun x => f x = true).
  Proof.
    split.
    intros x Hx. 
    unfold carrier in Hx; apply filter_In in Hx.
    destruct Hx; auto.
    intros x Hx.
    apply filter_In.
    pred_solve.
  Qed.

  Hint Immediate carrier_filter.

  Proposition carrier_inter (l1 l2 : list X) : carrier l1 cap carrier l2 ~ carrier (filter (carrier_bool l2) l1).
  Proof.
    apply id_transitive with (B := carrier l1 cap (fun x => carrier_bool l2 x = true)); auto.
  Qed.

  Variable P : X -> Prop.
  Hypothesis P_oracle : forall x, P x \/ ~ P x.

  Proposition carrier_select_dec l : carrier l inc P \/ exists x, carrier l x /\ ~ P x.
  Proof.
    induction l.
    left; auto. 
    destruct IHl as [ IHl | [ b [ H1 H2 ] ] ].
    destruct (P_oracle a).
    left; auto.
    right; exists a; split; auto.
    right; exists b; split; auto.
  Qed.
      
End Carrier.

Hint Resolve carrier_nil_1.
Hint Immediate carrier_nil_2.
Hint Immediate carrier_nil.
Hint Immediate carrier_cons_hd.
Hint Resolve carrier_cons_tail.
Hint Resolve carrier_cons_2.
Hint Resolve carrier_app.
Hint Resolve carrier_app_2.
Hint Immediate carrier_app_lft carrier_app_rt.
Hint Immediate carrier_filter.
Hint Extern 0 ( carrier ?l inc carrier (_::?l) )=> intro; apply carrier_cons_tail. 

