(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.

Require Import predicate.
Require Import carrier.
Require Import finite.

Section Constraint_closure.

  Set Implicit Arguments.

  Variable   X : Type.

  Variable   cl : Predicate X -> Predicate X.

  Definition Closure_increasing := forall A, A inc cl A.
  Definition Closure_monotonic  := forall A B, A inc B -> cl A inc cl B.
  Definition Closure_idempotent := forall A, cl (cl A) inc cl A.
  Definition Closure_compact    := forall C x, cl C x -> exists Cf, finite Cf /\ Cf inc C /\ cl Cf x.

  Hypothesis cl_monotonic  : Closure_monotonic.
  Hypothesis cl_increasing : Closure_increasing.
  Hypothesis cl_idempotent : Closure_idempotent.
  Hypothesis cl_cpct       : Closure_compact.

  Set Implicit Arguments.

  Proposition cl_inc C D : C inc cl D -> cl C inc cl D.
  Proof. intros ?; apply inc_transitive with (2 := cl_idempotent (A := D)); auto. Qed.

  Proposition inc_cl C D : cl C inc D -> C inc D.
  Proof. intros H; apply inc_transitive with (2 := H); auto. Qed.

  Hint Resolve cl_inc inc_cl.

  Proposition cl_closed C : cl (cl C) inc cl C.
  Proof. auto. Qed.

  Proposition cl_fixpoint C : cl (cl C) ~ cl C.
  Proof. split; auto. Qed.

  Proposition cl_closed_id C : cl C inc C -> cl C ~ C.
  Proof. split; auto. Qed.

  Proposition cl_id C D : C ~ D -> cl C ~ cl D.
  Proof. intros [ ]; split; auto. Qed.

  Proposition cl_inc_closed C D : C inc D -> cl D inc D -> cl C inc D.
  Proof. intros ? H; apply inc_transitive with (2 := H); auto. Qed.
 
  Hint Immediate cl_fixpoint cl_closed.
  Hint Resolve cl_id cl_inc_closed cl_closed_id.

  Let cl_extend_id_Rr C C' R : cl C ~ cl C' -> cl (C cup R) ~ cl (C' cup R).
  Proof.
    intros [ H1 H2 ]; apply inc_cl in H1; apply inc_cl in H2.
    split; apply cl_inc; apply bin_union_inc_l; auto.
    apply inc_transitive with (1 := H1); auto.
    apply inc_transitive with (1 := H2); auto.
  Qed.

  Proposition cl_extend_id_sg C C' c : cl C ~ cl C' -> cl (C cup sg c) ~ cl (C' cup sg c).
  Proof. apply cl_extend_id_Rr. Qed.

  Proposition cl_extend_id C C' D D' : cl C ~ cl C' -> cl D ~ cl D' -> cl (C cup D) ~ cl (C' cup D').
  Proof.
    intros.
    apply id_transitive with (1 := cl_extend_id_Rr D H).
    apply id_transitive with (1 := cl_id (bin_union_comm_id _ _ _)).
    apply id_transitive with (2 := cl_id (bin_union_comm_id _ _ _)).
    apply cl_extend_id_Rr.
    auto.
  Qed.
    
  Section Compactness.
  
    Let cl_compact_rec l0 : forall R, carrier l0 inc cl R -> exists R', finite R' /\ R' inc R /\ carrier l0 inc cl R'.
    Proof.
      induction l0 as [ | c l0 ]; intros R Hl0.
      exists empty; auto.  
      destruct (cl_cpct (C := R) (x := c)) as [ R' [ Hf' [ Hinc' Hcl' ] ] ].
      apply Hl0; auto.
      destruct (IHl0 R) as [ R'' [ Hf'' [ Hinc'' Hcl'' ] ] ]. 
      intro; auto.
      exists (R' cup R'').
      split; auto.
      split; auto.
      apply carrier_cons_2.
      apply inc_in_transitive with (1 := Hcl'). 
      intro; apply cl_monotonic; auto.
      apply inc_transitive with (1 := Hcl'').
      auto.
    Qed.

    Proposition cl_compact Cf R : finite Cf -> Cf inc cl R -> exists Rf, finite Rf /\ Rf inc R /\ Cf inc cl Rf.
    Proof.
      intros [ l0 [ Hl1 Hl2 ] ] Hinc.
      destruct (cl_compact_rec (l0 :=l0) (R:=R)) as [ R' [ Hf' [ Hinc' Hcl' ] ] ].
      apply inc_transitive with (1:= Hl1); auto.
      exists R'. 
      repeat (split; auto). 
      apply inc_transitive with (1:= Hl2); auto.
    Qed.

  End Compactness.

  Proposition cl_cl_cup_simplify R S : cl (R cup S) ~ cl (cl R cup S).
  Proof. split; auto. Qed.
 
  Proposition cl_extend_id_sg' A B C c : cl A ~ cl B -> B cup (sg c) ~ C -> cl (A cup sg c) ~ cl C.
  Proof.
    intros ? H2.
    apply id_transitive with (2 := cl_id H2).
    apply cl_extend_id_Rr; auto.
  Qed.

  Definition closed_lifted {K} (phi : K -> Predicate X) := forall x, cl (phi x) inc phi x.

  Proposition inter_closed K (phi : K -> Predicate X): closed_lifted phi -> cl (inter phi) inc inter phi.
  Proof.
    intros Hphi. 
    apply inter_inc_r.
    intros x; apply inc_transitive with (2 := Hphi x), cl_monotonic, inter_inc_l.
  Qed.

  Proposition cl_union_closed K (phi : K -> Predicate X) : cl (cl (union phi)) inc cl (union phi).
  Proof. apply cl_closed. Qed.

  Theorem cl_union_lub K (phi : K -> Predicate X) A : cl A inc A -> ( cl (union phi) inc A <-> forall x, phi x inc A ).
  Proof.
    intros HA. split.
    intros H1 ?; apply inc_transitive with (2 := H1); auto.
    intros H1; apply inc_transitive with (2 := HA), cl_monotonic, union_inc_l; auto.
  Qed.

  Proposition closed_bin_inter A B : cl A inc A -> cl B inc B -> cl (A cap B) inc A cap B.
  Proof. intros; auto. Qed.

  Proposition cl_bin_inter A B : cl A inc A -> cl B inc B -> cl (A cap B) ~ A cap B.
  Proof. split; auto. Qed.
 
  Proposition closed_bin_cl_union A B : cl (cl (A cup B)) inc cl (A cup B).
  Proof. auto. Qed.

  Proposition bin_cl_union_inc_r1 A B : A inc cl (A cup B).
  Proof. auto. Qed.

  Proposition bin_cl_union_inc_r2 A B : B inc cl (A cup B).
  Proof. auto. Qed.

  Proposition bin_cl_union_inc_r A B C : cl (A cup B) inc C -> (A inc C /\ B inc C).
  Proof. intros H; split; apply inc_transitive with (2 := H); auto. Qed.
 
  Proposition bin_cl_union_inc_l A B C : cl C inc C -> A inc C -> B inc C -> cl (A cup B) inc C.
  Proof. intros HC ? ?; apply inc_transitive with (2 := HC); auto. Qed.

  Hint Immediate closed_bin_cl_union bin_cl_union_inc_r1 bin_cl_union_inc_r2.
  Hint Resolve closed_bin_inter bin_cl_union_inc_r bin_cl_union_inc_l.

  Proposition bin_cl_union_lub A B C : cl C inc C -> ( cl (A cup B) inc C <-> A inc C /\ B inc C ).
  Proof. split; auto; intros []; auto. Qed.

End Constraint_closure.

Hint Resolve cl_inc inc_cl.
Hint Immediate cl_fixpoint cl_closed.
Hint Resolve cl_id cl_inc_closed cl_closed_id.
Hint Immediate closed_bin_cl_union bin_cl_union_inc_r1 bin_cl_union_inc_r2.
Hint Resolve closed_bin_inter bin_cl_union_inc_r bin_cl_union_inc_l.

