(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

(* This code is a rewrite from code of Jean Duprat and Damien Pous *)

Require Import Arith.
Require Import Omega.

Require Import replace.

Set Implicit Arguments.

Section dep_list.

  Variable X : Type.

  Inductive dlist : nat -> Type :=
    | dl_nil  : dlist 0
    | dl_cons : forall n, X -> dlist n -> dlist (S n).
  
  Definition dl_head n (l : dlist (S n)) : X :=
    match l in dlist k return 0 < k -> X with 
    | dl_nil        => fun h => False_rect _ (lt_n_O _ h)
    | dl_cons _ a _ => fun _ => a 
    end (lt_O_Sn n).
  
  Definition dl_tail n (l : dlist (S n)) : dlist n :=
    match l in dlist k return 0 < k -> dlist (pred k) with 
    | dl_nil        => fun h => False_rect (dlist (pred 0)) (lt_n_O _ h)
    | dl_cons _ _ q => fun _ => q
    end (lt_O_Sn n).
  
  Fixpoint dl_get n (l : dlist n) k {struct l} : k < n -> X :=
    match l in dlist n return k < n -> X with
    | dl_nil          => fun h => False_rect _ (lt_n_O _ h)
    | dl_cons n a q => match k return k < S n -> X with 
                | 0   => fun _ => a
  	        | S k => fun h => dl_get q (lt_S_n _ _ h)
  	    end
    end.

  Implicit Arguments dl_get [n].

  Let get_eq_rec n (l : dlist n) : forall x (H H' : x < n), dl_get l x H = dl_get l x H'.
  Proof.
    induction l; intros a H H'.
    contradict H; omega.
    unfold dl_get; fold dl_get.
    induction a; trivial.
  Qed.

  Fact dl_get_eq n l x y (Hx : x < n) (Hy : y < n) : x = y -> dl_get l x Hx = dl_get l y Hy.
  Proof. intros; subst; apply get_eq_rec. Qed.

  Fact dl_get_cons n a l : forall i (H : S i < S n) (H' : i < n),  dl_get (dl_cons a l) (S i) H = dl_get l i H'.
  Proof.
    destruct l as [ | n a' q ]; intros i H.
    contradict H; omega.
    simpl; induction i; trivial.
  Qed.

  Fixpoint dl_change n (l : dlist n) i (x : X) {struct l}: dlist n :=
    match l in dlist n, i return dlist n with
      | dl_nil          , _     => dl_nil
      | dl_cons _ a' q  , 0     => dl_cons x q
      | dl_cons _ a' q  , S i   => dl_cons a' (dl_change q i x)	
  end.	

  Fact dl_get_change_ii n a l : forall i (H : i < n), dl_get (dl_change l i a) i H = a.
  Proof.
    induction l; intros i H.
    contradict H; omega.
    induction i; simpl; trivial.
  Qed.

  Fact dl_get_change_ij: forall n a l i j (D: i <> j) (H : j < n), dl_get (dl_change l i a) j H = dl_get l j H.
  Proof.   
    induction l; intros i j D H.
    contradict H; omega.
    induction i; induction j; trivial.
    contradict D; trivial.
    simpl; apply IHl; intros H'; apply D; rewrite H'; trivial.
  Qed.

  Fact dl_0_nil : forall l : dlist 0, l = dl_nil.
  Proof.
    refine ((dlist_ind (fun n => match n return dlist n -> Prop with 
                                   | 0 => fun l => l = dl_nil 
                                   | _ => fun _ => True 
                                 end) (refl_equal dl_nil) _) 0);
    auto.  
  Qed.
    
  Fact dl_n0_cons : forall n (l : dlist (S n)), exists x, exists q, l = dl_cons x q.
  Proof.
    intro n.
    refine ((dlist_ind (fun n => match n return dlist n -> Prop with 
                                   | S n => fun l => exists x, exists q, l = dl_cons x q 
                                   | _   => fun _ => True 
                                 end) _ _) (S n));
    trivial.  
    intros m m' l H; exists m'; exists l; reflexivity.
  Qed.

  Fact dl_head_tail n (l : dlist (S n)) : l = dl_cons (dl_head l) (dl_tail l).
  Proof.
    generalize (dl_n0_cons l); intro H1; elim H1; intros x H2; elim H2; intros q H; clear H1 H2.
    rewrite H; simpl; reflexivity.
  Qed.

  Fixpoint dl_rev_init n (f : nat -> X) {struct n}: dlist n :=
    match n return dlist n with
    | 0   => dl_nil
    | S m => dl_cons (f n) (dl_rev_init m f)
    end.

  Definition dl_init n f := dl_rev_init n (fun m => f (n-m)).

  Fact dl_get_rev_init f n : forall i (H : i < n), dl_get (dl_rev_init n f) _ H = f (n-i).
  Proof.
    induction n; intros i H.
    contradict H; omega.
    induction i; simpl; trivial.
  Qed.

  Fact dl_get_init f n i (H : i < n) : dl_get (dl_init n f) i H = f i.
  Proof.
    unfold dl_init; intros; rewrite dl_get_rev_init.
    f_equal. omega.
  Qed.

  Fact dl_dlist_eq n : forall (l l': dlist n),  (forall i (H : i < n), (dl_get l i H = dl_get l' i H)) -> l = l'.
  Proof.
    induction l as [ | n x q IHl ]; intros l' H.
    rewrite (dl_0_nil l'); trivial.
    rewrite (dl_head_tail l') in H |- *. 
    f_equal. assert (0 < S n) as Hn. omega. 
    apply (H 0 Hn).
    apply IHl.
    intros i H0.
    assert (S i < S n) as H2. omega.
    generalize (H (S i) H2). clear H. intros H3.
    simpl in H3.
    generalize (get_eq_rec q H0 (lt_S_n _ _ H2)). intros H4.
    rewrite H4, H3. apply get_eq_rec.
  Qed.

  Fixpoint dl_repeat n x := match n return dlist n with 
    | 0   => dl_nil
    | S n => dl_cons x (dl_repeat n x)
  end.

  Proposition dl_repeat_prop x : forall n i (Hi : i < n), dl_get (dl_repeat n x) _ Hi = x.
  Proof.
    induction n; intros [ | i ]  Hi;
    try (exfalso; omega);
    simpl dl_repeat; auto.
    simpl.
    apply (IHn _ (lt_S_n _ _ Hi)).
  Qed.

  Section dlist_fall.

    Variable P : X -> Prop.
  
    Definition dl_fall n (l : dlist n) := forall x H, P (dl_get l x H).

    Theorem dl_fall_nil : dl_fall dl_nil.
    Proof. intros ? ?; exfalso; omega. Qed. 

    Theorem dl_fall_cons n x (l : dlist n) : dl_fall (dl_cons x l) <->  (P x /\ dl_fall l).
    Proof.
      split.
      
      intros H1; split. 
      generalize (H1 _ (lt_O_Sn _)); auto.
      intros i Hi.
      generalize (lt_n_S _ _ Hi); intros Hx'.
      generalize (H1 _ Hx').
      simpl.
      intros H; replace_with H. 
      intros [ H1 H2 ] [ | y ] Hy; simpl; auto.
    Qed.

  End dlist_fall.

  Hint Resolve dl_fall_nil dl_fall_cons.

  Section dl_fall_t.

    Variable f : X -> Type.
     
    Inductive dl_fall_t : forall n, dlist n -> Type := 
      | in_dl_fall_t_0 : dl_fall_t dl_nil
      | in_dl_fall_t_1 : forall n x (v : dlist n), f x -> dl_fall_t v -> dl_fall_t (dl_cons x v).

    Proposition dl_fall_t_get n (v : dlist n) : dl_fall_t v -> forall x (H : x < n), f (dl_get v _ H).
    Proof. 
      induction 1; intros [ | i ] Hi; try (exfalso; omega); simpl; auto.
    Qed.
  
    Proposition dl_get_fall_t n (v : dlist n) : (forall x (H : x < n), f (dl_get v _ H)) -> dl_fall_t v.
    Proof.
      intros H.
      induction v.
      constructor 1.
      constructor 2.
      apply (H _ (lt_O_Sn _)).
      apply IHv.
      intros i Hi.
      generalize (H _ (lt_n_S _ _ Hi)); intros H'.
      replace_with H'.
      simpl; auto.
    Qed.

    (* Beware that the maps dl_fall_t_get and dl_get_fall_t are NOT inverse to one another 
       
       For every x and ONE proof of Hx :x < n, the type dl_fall_t v contains an element in f (dl_get v x Hx)
       For every x and EVERY proof of Hx : x < n, the type (forall x (H : x < n), f (dl_get v x Hx)) contains an element in f (dl_get v x Hx)
         this element may have different values depending on the proof Hx 
    *)

    Proposition dl_fall_t_head_tail n (v : dlist (S n)) : dl_fall_t v -> f (dl_head v) * dl_fall_t (dl_tail v).
    Proof.
      intros H.
      rewrite dl_head_tail in H.
      generalize (dl_fall_t_get H); clear H; intros H.
      split.
      generalize (H _ (lt_O_Sn n)); auto.      
      apply dl_get_fall_t.
      intros x Hx.
      generalize (H _ (lt_n_S _ _ Hx)).
      simpl; intros H'; replace_with H'. 
    Qed.

    Proposition dl_fall_t_nil (v : dlist 0) : dl_fall_t v.
    Proof. rewrite dl_0_nil; apply in_dl_fall_t_0. Qed.

    Proposition dl_fall_t_cons n x (v : dlist n) : dl_fall_t (dl_cons x v) -> f x * dl_fall_t v.
    Proof.
      intros H.
      generalize (dl_fall_t_head_tail H).
      simpl; auto.
    Qed.

  End dl_fall_t.

  Theorem dl_fall_t_fall (P : X -> Prop) n (v : dlist n) : dl_fall_t P v -> dl_fall P v.
  Proof.
    intros H.
    generalize (dl_fall_t_get H).
    intros H' x Hx; apply H'.
  Qed.

  Theorem dl_fall_fall_t (P : X -> Prop) n (v : dlist n) : dl_fall P v -> dl_fall_t P v.
  Proof. intros; apply  dl_get_fall_t; auto. Qed.

End dep_list.

Hint Resolve dl_fall_nil dl_fall_t_fall dl_fall_fall_t.

Section dl_fall_t.

  Variable X Y : Type.

  Fixpoint dl_fall_t_dlist n (v : dlist X n) (fa : dl_fall_t (fun (_ : X) => Y) v) : dlist Y n :=
    match fa in @dl_fall_t _ _ n v return dlist Y n with
      | in_dl_fall_t_0 => @dl_nil Y
      | in_dl_fall_t_1 n x v' y fa' => dl_cons y (dl_fall_t_dlist fa')
    end.

End dl_fall_t.

Implicit Arguments dl_get [X n].

Section dep_list_map.

  Variables A B: Type.

  Section def.
 
    Variable f : A -> B.
    
    Fixpoint dl_map n (l : dlist A n) {struct l} : dlist B n :=
      match l in dlist _ n return dlist _ n with
        | dl_nil        => dl_nil B
        | dl_cons _ a q => dl_cons (f a) (dl_map q)
      end.

  End def.

  Implicit Arguments dl_map [n].

  Fact dl_get_map f n l : forall i (H : i < n), dl_get (dl_map f l) i H = f (dl_get l i H).
  Proof.
    induction l; intros i H.
    contradict H; omega.
    induction i; simpl; trivial.
  Qed.

  Fact dl_map_get_eq f g n l : (forall i (H : i < n), f (dl_get l _ H) = g (dl_get l _ H)) -> dl_map f l = dl_map g l.
  Proof.
    induction l; intros H.
    trivial.
    simpl. f_equal.
    assert (0 < S n) as H'. omega.
    apply (H _ H').
    apply IHl.
    intros i H'.
    assert (S i < S n) as H''. omega.
    replace_with (H _ H''); simpl;
    apply dl_get_eq; auto.
  Qed.

End dep_list_map.

Implicit Arguments dl_get_map [ A B n ].

Section dep_list_instanciate.

  Variable A B : Type.

  Fixpoint dl_inst n (l : dlist (A->B) n) (a : A) { struct l } : dlist B n :=
    match l in dlist _ n return dlist _ n with
        | dl_nil        => dl_nil B
        | dl_cons _ f q => dl_cons (f a) (dl_inst q a)
      end.

  Fact dl_get_inst n l : forall a i (H : i < n), dl_get (dl_inst l a) i H = dl_get l i H a.
  Proof.
    induction l; intros a i H.
    contradict H; omega.
    induction i; simpl; trivial.
  Qed.

End dep_list_instanciate.

Require Import Max.

Section dep_list_max.

  Fixpoint dl_max n l : nat :=
    match l with
      | dl_nil        => 0
      | dl_cons _ a l => max a (dl_max _ l)
    end.

  Fact dl_get_max n l : forall i (H : i < n), dl_get l i H <= dl_max l.
  Proof.
   induction l; intros i H.
   contradict H; omega.
   induction i; simpl.
   apply le_max_l.
   apply le_trans with (2 := le_max_r _ _).
   apply IHl.
  Qed.

  Fact dl_max_eq_n0 n (l : dlist _ n) : 0 < n -> { i : nat & { H : i < n | dl_get l i H = dl_max l }}.
  Proof.
    induction l; intros Hn.
    contradict Hn; omega.
    induction n.
    exists 0. exists (le_refl _).
    rewrite (dl_0_nil l); simpl; rewrite max_l; omega.
    clear IHn. destruct IHl as [ i [ Hi IH ] ]; [ omega | idtac ].
    destruct (max_dec x (dl_max l)) as [ H | H ].
    exists 0. exists Hn. simpl. omega.
    exists (S i). exists (lt_n_S _ _ Hi).
    unfold dl_max; fold dl_max.
    rewrite dl_get_cons with (H' := Hi), IH. auto.
  Qed.

  Fact dl_max_eq_0 (l : dlist _ 0) : dl_max l = 0.
  Proof. rewrite (dl_0_nil l); trivial. Qed.

  Fact dl_max_lub n : forall (l : dlist _ n) m, (forall x Hx, dl_get l x Hx <= m) -> dl_max l <= m.
  Proof.
    induction n; intros l m H.
    rewrite dl_max_eq_0. omega.
    assert (0 < S n) as Hn. omega.
    destruct (dl_max_eq_n0 l Hn) as [ i [ Hi H' ] ].
    rewrite <- H'. apply H.
  Qed.

End dep_list_max.

Section dep_list_sum.

  Fixpoint dl_sum n l : nat :=
    match l with
      | dl_nil        => 0
      | dl_cons _ a l => a+(dl_sum _ l)
    end.

End dep_list_sum.

(* notations *)

Implicit Arguments dl_nil [ X ].

Infix "##" := dl_cons (at level 60, right associativity).



