(**************************************************************)
(*   The Strong Completeness of Partial Monoidal BI           *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import MinMax.

Require Import predicate.

Require Import dlist.
Require Import term.

Require Import enum.
Require Import enumerations.

Parameter Var : Type.
Parameter enum_var : enumeration Var.

Inductive BI_Cst : Type := BI_Unit | BI_Bot.
Inductive BI_BinOp : Type := BI_Star | BI_Wand | BI_And | BI_Or | BI_Imp.

Inductive formula : Type :=
  | Form_var (_ : Var)
  | Form_cst (_ : BI_Cst)
  | Form_binop (_ : BI_BinOp) (_ _ : formula).

Section FormEnum.

  Let cst_enum : enumeration BI_Cst.
  Proof.
    exists (fun n => match n with 0 => BI_Unit | _ => BI_Bot end)
           (fun x => match x with BI_Unit => 0 | _ => 1 end);
    induction x; auto.
  Qed.

  Let binop_enum : enumeration BI_BinOp.
  Proof.
    exists (fun n => match n with | 0 => BI_Star | 1 => BI_Wand | 2 => BI_And | 3 => BI_Or | _ => BI_Imp end)
           (fun x => match x with | BI_Star => 0 | BI_Wand => 1 | BI_And => 2 | BI_Or => 3 | BI_Imp => 4 end);
    induction x; auto.
  Qed.

  Let sig n : Type := 
    match n with 
      | 0 => (Var + BI_Cst)%type
      | 2 => BI_BinOp
      | _ => Empty_set
    end.

  Let f n := match n as k return sig k -> dlist formula k -> formula with
               | 0 => fun x  _ => match x with inl v => Form_var v | inr c => Form_cst c end 
               | 2 => fun o v  => Form_binop o (dl_head v) (dl_head (dl_tail v))
               | _ => fun c  _ => match c with end
             end.

  Let term_form := term_recX f.

  Let term_form_prop n (c : sig n) v : term_form (c[v]) = f n c (dl_map term_form v).
  Proof. apply term_recX_prop. Qed.

  Let form_term : formula -> term sig :=
    (fix loop t := 
    match t return term sig with 
      | Form_var v         => @in_term sig 0 (inl _ v) dl_nil
      | Form_cst c         => @in_term sig 0 (inr _ c) dl_nil
      | Form_binop o f1 f2 => @in_term sig 2 o (loop f1##loop f2##dl_nil)
    end).

  Let term_form_eq t : term_form (form_term t) = t.
  Proof.
    induction t; simpl; auto.
    rewrite IHt1, IHt2; auto.
  Qed.

  Let form_term_eq t : form_term (term_form t) = t.
  Proof.
    induction t as [ n c v Hv ].
    rewrite term_form_prop.
    destruct n.
    destruct c; simpl; rewrite (dl_0_nil v); auto.
    destruct n.
    destruct c.
    destruct n.
    simpl.
    rewrite dl_head_tail in Hv.
    rewrite (dl_head_tail (dl_tail v)) in Hv.
    apply dl_fall_cons in Hv.
    destruct Hv as [ H1 H2 ].
    apply dl_fall_cons in H2.
    destruct H2 as [ H2 _ ].
    rewrite (dl_head_tail v), (dl_head_tail (dl_tail v)).
    simpl dl_map.
    simpl dl_head.
    rewrite H1, H2, (dl_0_nil (dl_tail _)); auto.
    destruct c.
  Qed.

  (* we have a bijection between terms over sig and bin_tree *)

  Let form_term_bij : bijection (term sig) formula.
  Proof. exists term_form form_term; auto. Qed.

  Hint Resolve unit_senum empty_senum enum_var.

  Theorem enum_formula : enumeration formula.
  Proof.
    apply enum_from_bij with (1:= form_term_bij).    
    apply term_enum.
    destruct n.
    apply enum_senum; auto.
    apply enum_sum; auto.
    destruct n; auto.
    destruct n; auto.
    apply (inr _ BI_Unit).
  Qed.

End FormEnum.

(* equality is decided using the enumeration *)

Definition form_eq_dec : forall x y : formula, { x = y } + { x <> y }.
Proof.
  apply enumeration_dec.
  apply enum_formula.
Qed.

Delimit Scope Formula_scope with formula.
  
Notation "x * y" := (Form_binop BI_Star x y) (at level 40, left associativity) : Formula_scope.
Notation "x '-*' y" := (Form_binop BI_Wand x y) (at level 41, left associativity) : Formula_scope.
Notation "x '&&' y" := (Form_binop BI_And x y) (at level 40, left associativity) : Formula_scope.
Notation "x '%%' y" := (Form_binop BI_Or x y) (at level 43, left associativity) : Formula_scope.
Notation "x '#>' y" := (Form_binop BI_Imp x y) (at level 44, left associativity) : Formula_scope.
Notation "% v" := (Form_var v) (at level 30) : Formula_scope.
Notation I := (Form_cst BI_Unit).
Notation BB := (Form_cst BI_Bot).

Open Scope Formula_scope.
