(**************************************************************)
(*   The Strong Completeness of Partial Monoidal BI           *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.
Require Import Arith.

Require Import predicate.
Require Import enum.
Require Import carrier.
Require Import finite.
Require Import maps.
Require Import nathole.

(* we have an infinite and denumerable set of labels *)

Parameter label : Type.
Parameter idx_label : indexation label.

Definition l0 : label.
Proof. apply (idx_dmap _ idx_label 0). Qed.

(* we can find a new label given a finite set of labels *)

Proposition new_label (P : label -> Prop) : finite P -> exists a, ~ P a.
Proof.
  intros HP.
  set (d := idx_dmap _  idx_label).
  set (r := idx_rmap _ idx_label).
  set (Q := pd_image r P).
  assert (finite Q) as HQ.
  apply finite_pd_image. trivial.
  apply finite_nat_bounded in HQ.
  destruct HQ as [ k HQ ].
  exists (d (S k)).
  intros H.
  assert (Q (S k)) as H'.
  exists (d (S k)). split. trivial.
  apply (idx_inj). apply HQ in H'.
  apply (le_Sn_n _ H').
Qed.

(* and we can find two new labels *)

Proposition new_label2 (P : label -> Prop) : finite P -> exists a, exists b, a <> b /\ ~ P a /\ ~ P b.
Proof.
  intros HP.
  destruct (new_label _ HP) as [ a Ha ].
  set (Q := P cup sg a).
  assert (finite Q) as HQ. unfold Q; auto.
  destruct (new_label _ HQ) as [ b Hb ].
  exists a. exists b. split.
  intros H. apply Hb. rewrite <- H.
  unfold Q; pred_solve.
  split; auto.
  intros H; apply Hb.
  unfold Q; pred_solve.
Qed.

Definition label_eq_dec (a b : label) : { a = b } + { a <> b }.
Proof. apply enumeration_dec, enum_from_idx, idx_label; auto. Qed.

Hint Immediate label_eq_dec.

Section injection.

  Variables A B  : label -> Prop.
  Hypothesis HA  : finite A.
  Hypothesis HB  : finite B.
  Hypothesis HAB : A inc B.

  Proposition label_hole_injection : exists f : label -> label, injective f /\ forall x, (A x /\ f x = x) \/ (~ A x /\ ~ B (f x)).
  Proof.
    destruct idx_label as [ dmap rmap req deq ].
    set (A' := pd_image rmap A).
    set (B' := pd_image rmap B).
    assert (finite A') as HA'. apply finite_pd_image; auto.
    assert (finite B') as HB'. apply finite_pd_image; auto.
    assert (A' inc B') as HAB'.
      unfold A', B'.
      intros n [ x [ H1 H2 ] ].
      exists x; split; auto.
    destruct (nat_hole_injection _ _ HA' HB' HAB') as ( f & H1 & H2 ).
    exists (fun x => dmap (f (rmap x))).
    split.
    
    intros x1 x2 H.
    assert (rmap x1 = rmap x2) as H'.
      apply H1.
      rewrite <- req, <- H, req; auto.
    rewrite <- deq, <- H', deq; auto.
    
    intros x.
    destruct (H2 (rmap x)) as [ [ (y & H3 & H4) H5 ] | [ H3 H4 ] ].

    left. 
    split. 
    rewrite <- deq, <- H4, deq; auto.
    rewrite H5; auto.
    
    right.
    split.
    intros H; apply H3; exists x; auto.
    intros H; apply H4; exists (dmap (f (rmap x))); auto.
  Qed.

End injection.
