(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.
Require Import list_fall_exst.
Require Import list_in_t.

Section sigT_sig_iso.

  Set Implicit Arguments.
 
  Variable X : Type.
  Variable P : X -> Prop.

  Implicit Type xt : sigT P.
  Implicit Type x : sig P.

  Theorem sigT_sig x : sig_of_sigT (sigT_of_sig x) = x.
  Proof. induction x; auto. Qed.

  Theorem sig_sigT xt : sigT_of_sig (sig_of_sigT xt) = xt.
  Proof. induction xt; auto. Qed.

  Definition lsig_of_lsigT := map (sig_of_sigT (P:=P)).
  Definition lsigT_of_lsig := map (sigT_of_sig (P:=P)).

  Theorem lsigT_lsig l : lsig_of_lsigT (lsigT_of_lsig l) = l.
  Proof. induction l; simpl; f_equal; auto. apply sigT_sig. Qed.

  Theorem lsig_lsigT l : lsigT_of_lsig (lsig_of_lsigT l) = l.
  Proof. induction l; simpl; f_equal; auto. apply sig_sigT. Qed.

End sigT_sig_iso.

Section Forall_Exists_t.

  Set Implicit Arguments.
  
  Variable X : Type.

  Section def_props.

    Variable f : X -> Type.
  
    Inductive Forall_t : list X -> Type :=
      | in_Forall_t_0 : Forall_t nil
      | in_Forall_t_1 : forall x ll, f x -> Forall_t ll -> Forall_t (x::ll).

    Hint Constructors Forall_t.

    Theorem Forall_t_inv a l : Forall_t (a::l) -> (f a) * (Forall_t l).
    Proof. intros H; inversion_clear H; split; auto. Qed.

    Theorem Forall_t_app l r : Forall_t l -> Forall_t r -> Forall_t (l++r).
    Proof.
      induction 1; auto.
      simpl; intro;
      constructor; auto.
    Qed.

    Theorem Forall_t_app_inv l r : Forall_t (l++r) -> Forall_t l * Forall_t r.
    Proof.
      induction l; intros H.
      split; auto. 
      simpl in H.
      apply Forall_t_inv in H.
      destruct H as [ H1 H2 ].
      apply IHl in H2.
      destruct H2 as [ H2 H3 ].
      auto.
    Qed.

    Theorem Forall_t_imp_In_t l : Forall_t l -> forall x, In_t x l -> f x.
    Proof.
      induction 1. 
      intros x [ l [ r H ] ].
      induction l; discriminate H.
      intros y Hy.
      In_t_auto.
      rewrite Hy; auto.
    Qed.

    Theorem In_t_imp_Forall_t l : (forall x, In_t x l -> f x) -> Forall_t l.
    Proof.
      induction l.
      intro; constructor.
      intros H.
      constructor.
      apply H; In_t_auto.
      apply IHl; intros x H'; apply H; In_t_auto.
    Qed.

    Inductive Exists_t : list X -> Type :=
      | in_Exists_t_0 : forall x l, f x -> Exists_t (x::l)
      | in_Exists_t_1 : forall x l, Exists_t l -> Exists_t (x::l).

    Theorem Exists_t_app_l l r : Exists_t l -> Exists_t (l++r).
    Proof.
      induction 1.
      constructor 1; auto.
      simpl.
      constructor 2; auto.
    Qed.

    Theorem Exists_t_app_r l r : Exists_t r -> Exists_t (l++r).
    Proof.
      induction l; auto.
      intros Hr.
      constructor 2; auto.
    Qed.

    Theorem Exists_t_app l r : Exists_t (l++r) -> Exists_t l + Exists_t r.
    Proof.
     induction l; simpl.
     intros H; right; auto.
     inversion 1.
     left. constructor 1; auto.
     apply IHl in X1.
     destruct X1 as [ X1 | X1 ].
     left; constructor 2; auto.
     right; auto.
    Qed.

    Theorem Exists_t_imp_In_t : forall l, Exists_t l -> { x : X &  f x & In_t x l }.
    Proof. induction 1 as [ x l Hx |  x l Hx [ y H1 H2 ] ];  [ exists x | exists y ]; auto. Qed.

    Theorem In_t_imp_Exists_t ll : { x : X & f x & In_t x ll } -> Exists_t ll.
    Proof.
      intros [ x Tx [ l [ r Hlr ] ] ].
      rewrite Hlr. clear ll Hlr.
      apply Exists_t_app_r.
      constructor 1; auto.
    Qed. 

  End def_props.

  Section Forall_t_lsigT_iso.

    Variable f : X -> Type.

    Fixpoint Forall_t_lsigT ll (Hll : Forall_t f ll) : list (sigT f) :=
      match Hll with
        | in_Forall_t_0 => nil
        | in_Forall_t_1 x ll Hx Hll => (existT _ x Hx)::(Forall_t_lsigT Hll)
      end.

    Global Implicit Arguments Forall_t_lsigT [].

    Theorem Forall_t_lsigT_prop ll (Hll : Forall_t f ll) : map (projT1 (P:=f)) (Forall_t_lsigT ll Hll) = ll.
    Proof.
      induction Hll. auto.
      simpl. f_equal.
      auto.
    Qed.

    Definition sigfall_t_lsigT (sf : sigT (Forall_t f)) :=
      match sf with
        | existT ll Hll => Forall_t_lsigT ll Hll
      end.

    Fixpoint lsigT_sigfall_t (l : list (sigT f)) : sigT (Forall_t f) :=
      match l with
        | nil               => existT (Forall_t f) nil (in_Forall_t_0 f)
        | (existT x Hx)::l  => 
           match lsigT_sigfall_t l with
             | existT ll Hll => existT (Forall_t f) (x::ll) (in_Forall_t_1 x Hx Hll)
           end
      end.

    Theorem lsigT_sigfall_t_prop l : sigfall_t_lsigT (lsigT_sigfall_t l) = l.
    Proof.
      induction l as [ | [ x Hx ] l ].
      simpl; auto.
      simpl.
      destruct (lsigT_sigfall_t l).
      simpl in * |- *.
      f_equal; auto.
    Qed.

    Theorem sigfall_t_lsigT_prop sf : lsigT_sigfall_t (sigfall_t_lsigT sf) = sf.
    Proof.
      destruct sf as [ ll Hll ].
      induction Hll.
      simpl; auto.
      simpl in * |- *.
      rewrite IHHll.
      auto.
    Qed.

  End Forall_t_lsigT_iso.

  Theorem Forall_t_Forall (P : X -> Prop) l : Forall_t P l -> Forall P l.
  Proof. induction 1; constructor; auto. Qed.

  Theorem Forall_Forall_t (P : X -> Prop) l : Forall P l -> Forall_t P l.
  Proof.
    intros H1.
    induction l as [ | x l ].
    apply (in_Forall_t_0 P).
    apply list_fall_cons in H1.
    constructor; tauto.
  Qed.

  Section map_props.

    Variable f g : X -> Type.
    Hypothesis map : forall x, f x -> g x.

    Theorem Forall_t_map l : Forall_t f l -> Forall_t g l.
    Proof. induction 1; constructor; auto. Qed.
    
    Hypothesis Hfg : forall x, f x + g x.
    
    Theorem Forall_t_Exists_t l : Forall_t f l + Exists_t g l.
    Proof.
      induction l.
      left; constructor.
      destruct IHl as [ IHl | IHl ].
      destruct (Hfg a) as [ H | H ].
      left; constructor; auto.
      right; constructor 1; auto.
      right; constructor 2; auto.
    Qed.

  End map_props.

  Section cond_props.

    Variable P : X -> Prop.
    Variable f g : X -> Type.
    Hypothesis HPfg : forall x, P x -> f x + g x.
    
    Theorem Forall_t_Exists_t_cond l : Forall P l -> Forall_t f l + Exists_t g l.
    Proof.
      induction l; intros HP.
      left; constructor.
      apply list_fall_cons in HP.
      destruct HP as [ H1 H2 ].
      apply IHl in H2. clear IHl.
      destruct H2 as [ IHl | IHl ].
      destruct (HPfg H1) as [ H | H ].
      left; constructor; auto.
      right; constructor 1; auto.
      right; constructor 2; auto.
    Qed.

  End cond_props.

  Section relations.

    Variable Y : Type.
    Variable Q : X -> Y -> Prop.

    Definition Forall_R (xx : list X) (yy : list Y) := exists cc, xx = map (fst (B:=Y)) cc 
                                                               /\ yy = map (snd (B:=Y)) cc 
                                                               /\ Forall (fun c => match c with (x,y) => Q x y end) cc.

    Theorem Forall_R_nil_l yy : Forall_R nil yy -> yy = nil.
    Proof.
      intros ([ | c cc] & H1 & H2 & H3).
      rewrite H2; auto.
      discriminate H1.
    Qed.

    Theorem Forall_R_nil_r xx : Forall_R xx nil -> xx = nil.
    Proof.
      intros ([ | c cc] & H1 & H2 & H3).
      rewrite H1; auto.
      discriminate H2.
    Qed.

    Theorem Forall_R_nil : Forall_R nil nil.
    Proof. exists nil; auto. Qed.

    Theorem Forall_R_cons x y xx yy : Q x y -> Forall_R xx yy -> Forall_R (x::xx) (y::yy).
    Proof.
      intros H (cc & H1 & H2 & H3).
      exists ((x,y)::cc); simpl; rewrite H1, H2; auto.
    Qed.

    Theorem Forall_R_cons_inv x xx y yy : Forall_R (x::xx) (y::yy) -> Q x y /\ Forall_R xx yy.
    Proof.
      intros ([ | (c1,c2) cc] & H1 & H2 & H3).
      discriminate H1.
      injection H1; clear H1; intros Hx Hxx.
      injection H2; clear H2; intros Hy Hyy.
      apply list_fall_cons in H3.
      destruct H3 as [ H1 H2 ].
      split.
      rewrite Hxx, Hyy; auto.
      exists cc; auto.
    Qed.

    Theorem Forall_R_cons_l x xx yy : Forall_R (x::xx) yy -> { y : Y & { yy' | Q x y /\ yy = y::yy' /\ Forall_R xx yy' } }.
    Proof.
      intros H. destruct yy as [ | y yy' ].
      apply Forall_R_nil_r in H; discriminate H.
      apply Forall_R_cons_inv in H.
      destruct H as [ H1 H2 ].
      exists y; exists yy'; auto.
    Qed.

    Theorem Forall_R_cons_r xx y yy : Forall_R xx (y::yy) -> { x : X & { xx' | Q x y /\ xx = x::xx' /\ Forall_R xx' yy } }.
    Proof.
      intros H. destruct xx as [ | x xx' ].
      apply Forall_R_nil_l in H; discriminate H.
      apply Forall_R_cons_inv in H.
      destruct H as [ H1 H2 ].
      exists x; exists xx'; auto.
    Qed.

    Theorem Forall_R_In_t x xx yy : Forall_R xx yy -> In_t x xx -> { y : Y & In_t y yy & Q x y }.
    Proof. 
      generalize x yy; clear x yy; induction xx;
      intros x yy Hxy Hyy.
      contradiction (In_t_nil Hyy).
      apply In_t_cons in Hyy.
      apply Forall_R_cons_l in Hxy.
      destruct Hxy as ( y & yy' & H1 & H2 & H3 ).
      destruct Hyy as [ Hyy | Hyy ].
      exists y.
      rewrite H2; In_t_auto.
      rewrite Hyy; auto.
      apply IHxx with (x:=x) in H3; auto.
      destruct H3 as ( y' & l & r & H3 ).
      exists y'; auto.
      rewrite H2, H3; In_t_auto.
    Qed.
      
    Theorem Forall_t_rel2list (l : list X) : Forall_t (fun x => { y | Q x y }) l -> { ll | Forall (fun c : X*Y => let (x,y):=c in Q x y) ll 
  	                                                                           /\ l = map (fst (B:=Y)) ll }.
    Proof.
      induction 1 as [ | x ll [ y Hy ] Hll [ ll0 [ H1 H2 ] ]].
      exists nil; auto.
      exists ((x,y)::ll0); split; auto; rewrite H2; simpl; auto.
    Qed.

    Definition Forall_t_sig_to_list xx : Forall_t (fun x => sig (Q x)) xx -> { yy | Forall_R xx yy }.
    Proof.
      induction xx as [ | x xx ]; intros Hxx.
      exists nil. exists nil; auto.
      apply Forall_t_inv in Hxx.
      destruct Hxx as [ [y Hy] Hxx ].
      apply IHxx in Hxx.
      destruct Hxx as [ yy Hyy ].
      exists (y::yy).
      destruct Hyy as ( cc & H1 & H2 & H3 ).
      exists ((x,y)::cc).
      simpl. rewrite H1, H2.
      split; auto.
    Qed.

    Definition Forall_t_list_to_sig xx : { yy | Forall_R xx yy } -> Forall_t (fun x => sig (Q x)) xx.
    Proof.
      induction xx as [ | x xx ]; intros Hxx.
      constructor.
      destruct Hxx as [ yy Hyy ].
      apply Forall_R_cons_l in Hyy.
      destruct Hyy as (y & yy' & H1 & H2 & H3).
      constructor.
      exists y; auto.
      apply IHxx.
      exists yy'; auto.
    Qed.

  End relations.

End Forall_Exists_t.

Theorem Forall_t_map_Forall_t X Y (f : X -> Type) (g : Y -> X) l : Forall_t f (map g l) -> Forall_t (fun y => f (g y)) l.
Proof.
  induction l as [ | y l ]; intros H.
  constructor.
  apply Forall_t_inv in H.
  destruct H as [ H1 H2 ].
  apply IHl in H2.
  constructor; auto.
Qed.

Theorem Forall_t_Forall_t_map X Y (f : X -> Type) (g : Y -> X) l : Forall_t (fun y => f (g y)) l -> Forall_t f (map g l).
Proof.
  induction l as [ | y l ]; intros H.
  constructor.
  apply Forall_t_inv in H.
  destruct H as [ H1 H2 ].
  apply IHl in H2.
  constructor; auto.
Qed.

Section Forall_t_Forall_t_sig.

  Variable X : Type.
  Variable f g : X -> Type.

  Theorem Forall_t_Forall_t_lsigT ll (Hll : Forall_t g ll) : Forall_t (fun s => f (projT1 s)) (Forall_t_lsigT ll Hll) -> Forall_t f ll.
  Proof.
    induction Hll; intros H.
    constructor.
    apply Forall_t_inv in H.
    destruct H as [ H1 H2 ].
    constructor; auto.
  Qed.

End Forall_t_Forall_t_sig.

Hint Constructors Forall_t Exists_t.
Hint Resolve Forall_t_app Forall_t_map Exists_t_app_l Exists_t_app_r Forall_t_Forall.
Hint Immediate Forall_R_nil.
Hint Resolve Forall_R_cons.

Section Intersection.

  Variable X : Type.

  Definition list_inter_empty l1 l2 := forall x : X, In x l1 -> In x l2 -> False.

  Hypothesis Xdec : forall x y : X, { x = y } + { x <> y }.  

  Let Forall_or_split ll (x : X) : ( Forall (fun y => x <> y) ll ) + In_t x ll.
  Proof.
    set(P (k : X) := (fun y => x <> y) k).
    set(Q (k : X) := x = k).
    destruct (Forall_t_Exists_t P Q) with (l := ll) as [ H | H ].
    intros k. destruct (Xdec x k); [ right | left ]; red; auto.
    left; auto.
    right.
    apply Exists_t_imp_In_t in H.
    unfold Q in H.
    destruct H as [ y H1 H2 ];
    rewrite H1; auto.
  Qed.

  Theorem Inter_empty_or_split_both ll1 ll2 : list_inter_empty ll1 ll2 
                                            + { x : X & In_t x ll1 & In_t x ll2 }.
  Proof.
    unfold list_inter_empty.
    destruct (Forall_t_Exists_t _ _ (Forall_or_split ll1) ll2) as [ H | H ].
    left; intros x H1 H2.
    apply Forall_t_Forall in H.
    generalize (proj1 (Forall_forall _ _) H _ H2); clear H; intros H.
    generalize (proj1 (Forall_forall _ _) H _ H1); clear H; intros H.
    apply H; auto.
    right.
    apply Exists_t_imp_In_t.
    auto.
  Qed.

End Intersection.

Section Forall_t_inhabited.

  Variable X : Type.
  Variable P : X -> Type.

  Proposition Forall_t_inhabited ll : Forall (fun x => inhabited (P x)) ll -> inhabited (Forall_t P ll).
  Proof.
    induction ll; auto.
    intros H; list_fall_auto.
    apply IHll in H0.
    destruct H; destruct H0.
    constructor; auto.
  Qed.

End Forall_t_inhabited.

Ltac list_fall_t_auto := repeat 
   match goal with
     | [H: Forall_t _ nil |- _]        => clear H 
     | [H: Forall_t _ (?x::?l2) |- _]  => apply Forall_t_inv in H; destruct H
     | [H: Forall_t _ (?l1++?l2) |- _] => apply Forall_t_app_inv in H; destruct H
   end.

Ltac list_fall_t_simpl_one H := apply Forall_t_Forall in H; list_fall_simpl_one H.
Ltac list_fall_t_simpl := repeat
  match goal with
    | [H: Forall_t ?P ?l |- _ ] => list_fall_t_simpl_one H
  end.

