(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.
Require Import list_fall_exst.

Section In_t.

  Variable X : Type.

  Set Implicit Arguments.

  Definition In_t (x : X) sys := { l : list X & { r | sys=l++x::r } }.

  Theorem In_t_hd : forall x l, In_t x (x::l).
  Proof. intros x l; exists nil; exists l; auto. Qed.

  Hint Immediate In_t_hd.

  Theorem In_t_hd' : forall x y l, x = y -> In_t x (y::l).
  Proof. intros x y l H; rewrite H; auto. Qed.

  Hint Resolve In_t_hd'.

  Theorem In_t_tl : forall x y l, In_t x l -> In_t x (y::l).
  Proof. 
    intros x y l [ l1 [ r1 H ] ].
    exists (y::l1); exists r1; 
    rewrite H; auto.
  Qed.

  Theorem In_t_In : forall x ll, In_t x ll -> In x ll.
  Proof.
    intros x ll [ l [ r H ] ].
    rewrite H.
    apply in_or_app.
    right; red; auto.
  Qed.

  Theorem In_t_app_l : forall x l m, In_t x l -> In_t x (l++m).
  Proof.
    intros x l m [ l1 [ r1 H ] ].
    exists l1. exists (r1++m).
    rewrite H.
    apply app_ass.
  Qed.

  Theorem In_t_app_r : forall x l m, In_t x m -> In_t x (l++m).
  Proof.
    intros x l m [ l1 [ r1 H ] ].
    exists (l++l1). exists r1.
    rewrite H.
    symmetry.
    apply app_ass.
  Qed.

  Hint Resolve In_t_hd In_t_tl In_t_app_l In_t_app_r.

  Theorem In_t_nil : forall x, In_t x nil -> False.
  Proof.
    intros x [ l [ r Hx ] ].
    induction l;
    discriminate Hx.
  Qed.

  Theorem In_t_cons : forall x y l, In_t x (y::l) -> (x = y) + (In_t x l).
  Proof.
    intros x y l [ l1 [ l2 Hl ] ].
    induction l1.
    left. injection Hl. auto.
    right. clear IHl1.
    injection Hl. clear Hl. intros HL _.
    exists l1. exists l2. auto.
  Qed.

  Theorem In_t_1 : forall x y, In_t x (y::nil) -> x = y.
  Proof.
    intros x y H.
    apply In_t_cons in H.
    destruct H as [ | H ]; auto.
    apply In_t_nil in H.
    contradiction H.
  Qed.

  Theorem In_t_2 : forall x y z, In_t x (y::z::nil) -> { x = y } + { x = z }.
  Proof.
    intros x y z H.
    apply In_t_cons in H.
    destruct H as [ | H ].
    left; auto.
    right.
    apply In_t_1 in H.
    auto.
  Qed.

  Theorem In_t_3 : forall x a b c ll, In_t x (a::b::c::ll) -> { x = a } + { x = b } + { x = c } + In_t x ll.
  Proof.
    intros x a b c ll H.
    apply In_t_cons in H.
    destruct H as [ | H ].
    left; auto.
    apply In_t_cons in H.
    destruct H as [ | H ].
    left; auto.
    apply In_t_cons in H.
    destruct H as [ | H ].
    left; auto.
    right; auto.
  Qed.

  Theorem In_t_split : forall x l m, In_t x (l++m) -> (In_t x l) + (In_t x m). 
  Proof.
    intros x l m.
    induction l; intros H.
    right; auto.
    simpl in H.
    apply In_t_cons in H.
    destruct H as [ H | H ].
    left. auto.
    apply IHl in H.
    destruct H as [ H | H ]; [ left | right ] ; auto.
  Qed.

  Definition In_t_inc l m := forall x, In_t x l -> In_t x m.

  Theorem In_t_inc_refl l : In_t_inc l l.
  Proof. intros; red; auto. Qed.

  Theorem In_t_inc_tran k l m  : In_t_inc k l -> In_t_inc l m -> In_t_inc k m.
  Proof. intros; red; auto. Qed.

  Theorem In_t_inc_cons a l : In_t_inc l (a::l).
  Proof. red; intros; auto. Qed.
  
  Theorem In_t_inc_left : forall l m, In_t_inc l (l++m).
  Proof. intros l m x; auto. Qed.

  Theorem In_t_inc_right : forall l m, In_t_inc m (l++m).
  Proof. intros l m x; auto. Qed.

  Hint Immediate In_t_inc_refl In_t_inc_cons In_t_inc_left In_t_inc_right.

  Theorem In_t_inc_left' : forall l1 l2 m, In_t_inc l1 l2 -> In_t_inc l1 (l2++m).
  Proof.
    intros l1 l2 m H.
    apply In_t_inc_tran with (1 := H); auto.
  Qed.

  Theorem In_t_inc_right' : forall l m1 m2, In_t_inc m1 m2 -> In_t_inc m1 (l++m2).
  Proof.
    intros l m1 m2 H.
    apply In_t_inc_tran with (1 := H); auto.
  Qed.

  Theorem In_t_inc_app : forall l m k, In_t_inc l k -> In_t_inc m k -> In_t_inc (l++m) k.
  Proof. 
    intros l m k H1 H2 x Hx.
    apply In_t_split in Hx.
    destruct Hx; auto.
  Qed.

  Hint Resolve In_t_inc_app.

  Theorem In_t_inc_exchg l m : In_t_inc (l++m) (m++l).
  Proof. intros; auto. Qed.

  Theorem In_t_inc_dup l : In_t_inc (l++l) l.
  Proof. intros; auto. Qed.

  Let In_t_inc_sub_rec a l r : In_t_inc a (l++a++r).
  Proof. 
    intros.
    apply In_t_inc_tran with (l := a++r); auto.
  Qed.
    
  Theorem In_t_inc_sub : forall a b l r, In_t_inc a b -> In_t_inc a (l++b++r).
  Proof.
    intros a b l r H.
    apply In_t_inc_tran with (1 := H); auto.
  Qed.

End In_t.

Hint Immediate In_t_inc_refl In_t_inc_cons In_t_inc_left In_t_inc_right In_t_inc_exchg In_t_inc_dup.
Hint Resolve In_t_hd In_t_tl In_t_In In_t_app_l In_t_app_r In_t_inc_app In_t_inc_sub In_t_inc_left' In_t_inc_right'.

Ltac tac_In_t_hyps := 
  match goal with 
    | [ H : In_t ?x nil      |- _ ] => contradiction (In_t_nil H)
    | [ H : In_t ?x (?y::?l) |- _ ] => apply In_t_cons in H; destruct H as [ H | H ]
    | [ H : In_t ?x (?l++?m) |- _ ] => apply In_t_split in H; destruct H as [ H | H ]
  end.

Ltac tac_In_t_goal := repeat (auto; 
  match goal with
    | [  |- In_t ?x (?y::?l) ] => try (apply In_t_hd; auto; fail); apply In_t_tl
    | [  |- In_t ?x (?l++?m) ] => try (apply In_t_app_l; tac_In_goal; fail); apply In_t_app_r
  end).

Ltac In_t_auto := repeat tac_In_t_hyps; try tac_In_t_goal.

Section In_t_map.

  Variable X Y : Type.
  Variable f : X -> Y.

  Lemma In_t_map : forall l x, In_t x l -> In_t (f x) (map f l).
  Proof.
    induction l.
    intros x H.
    contradiction (In_t_nil H).
    simpl; intros x Hx.
    apply In_t_cons in Hx.
    destruct Hx as [ Hx | Hx ].
    rewrite Hx; In_t_auto.
    apply IHl in Hx; In_t_auto.
  Qed.

  Lemma In_t_map_rev : forall l y, In_t y (map f l) -> { x : X & In_t x l & y = f x }.
  Proof.
    induction l; intros y Hy; simpl in Hy; In_t_auto.
    exists a; auto.
    apply IHl in Hy.
    destruct Hy as [ x H1 H2 ].
    exists x; auto.
  Qed.

End In_t_map.
    
Section find.

  Variable X : Type.
  Variable x : X.
  Variable eqx_dec : forall y : X, { x = y } + { x <> y }.

  Definition list_find (l : list X) : In x l -> In_t x l.
  Proof.
    induction l; intros H. 
    contradiction H.
    simpl in H.
    destruct (eqx_dec a) as [ H' | H' ].
    rewrite H'; auto.
    apply In_t_tl.
    apply IHl.
    destruct H as [ H | H ]; auto.
    symmetry in H.
    contradiction (H' H).
  Qed.

End find.

    
