(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import Omega.

Require Import enum.

Section Mult.

  Lemma mult_lt_reg_rec n m : n < m -> forall p, (S p)*n < (S p)*m.
  Proof. induction p; [ simpl | simpl in IHp |- * ] ; omega. Qed.

  Lemma mult_le_reg_l_S m n p : (S p) * n <= (S p) * m -> n <= m.
  Proof.
    intros H.
    destruct (le_or_lt n m); auto.
    contradict H.
    apply lt_not_le.
    apply mult_lt_reg_rec; auto.
  Qed.
 
  Proposition mult_le_reg_l m n p : 0 < p -> p * n <= p * m -> n <= m.
  Proof.
    elim p.
    intros H; contradict H; omega.
    clear p; intros p _ _.
    apply mult_le_reg_l_S.
  Qed.

  Proposition mult_reg_l m n p : 0 < p -> p * n = p * m -> n = m.
  Proof. intros Hp H; apply le_antisym; apply (mult_le_reg_l _ _ _ Hp); omega. Qed.

End Mult.

Section Induction. 

  Theorem nat_2_ind : forall (P : nat -> Prop), P 0 -> P 1 -> (forall n, P n -> P (S (S n))) -> forall n, P n.
  Proof.
    intros P H0 H1 Hr n.
    cut (P n /\ P (S n)).
    tauto.
    elim n; intuition.
  Qed.

  Theorem nat_gen_ind :  forall (P : nat -> Prop), (forall n, (forall k, k < n -> P k) -> P n) -> forall n, P n.
  Proof.
    intros P HP.
    set (Q n := forall k, k <= n -> P k).
    assert (forall n, Q n).
    induction n; intros k Hk; apply HP; intros r Hr. 
    contradict Hr; omega.
    apply IHn. omega.
    intros n; apply (H n); auto.
  Qed.

  Definition nat_gen_rec :  forall (P : nat -> Set), (forall n, (forall k, k < n -> P k) -> P n) -> forall n, P n.
  Proof.
    intros P HP.
    set (Q n := forall k, k <= n -> P k).
    assert (forall n, Q n).
    induction n; intros k Hk; apply HP; intros r Hr. 
    contradict Hr; omega.
    apply IHn. omega.
    intros n; apply (H n); auto.
  Qed.

End Induction.

Section div2_even.
   
  Fixpoint div2 n := match n with
    | 0 => 0
    | 1 => 0
    | S (S n) => S (div2 n)
  end.

  Proposition div2_2x : forall x, div2 (2*x) = x.
  Proof.
    apply nat_2_ind; try (simpl; auto; fail).
    intros n Hn.
    cutrewrite (2*S (S n) = S (S (S (S (2*n))))).
    unfold div2; fold div2. f_equal; f_equal; auto.
    omega.
  Qed.

  Inductive even : nat -> Prop :=
    | e0 : even 0
    | e1 : forall n, odd n -> even (S n)
  with odd : nat -> Prop :=
    | o1 : forall n, even n -> odd (S n).

  Hint Constructors even.
  Hint Constructors odd.

  Scheme even_indr := Induction for even Sort Prop
     with     odd_indr := Induction for odd Sort Prop.

  Theorem even_or_odd : forall n, { even n } + { odd n }.
  Proof.
    induction n.
    left; apply e0.
    destruct IHn.
    right; apply o1; trivial.
    left; apply e1; trivial.
  Qed.

  Theorem even_2x : forall x, even (2*x).
  Proof.
    induction x.
    auto. 
    cutrewrite (2*(S x) = S (S (2*x))).
    auto.
    omega.
  Qed.

  Theorem odd_2x1 : forall x, odd (2*x+1).
  Proof.
    induction x. simpl.
    auto. 
    cutrewrite (2*(S x) + 1 = S (S (S (2*x)))); [ | omega ].
    apply o1, e1.
    rewrite plus_comm in IHx; auto.
  Qed. 

  Theorem not_even_odd : forall n, even n -> odd n -> False.
  Proof.
    induction n.
    intros E0 O0. inversion O0.
    intros ESn OSn. inversion ESn. inversion OSn. auto.
  Qed.

  Let div2_lemma_rec : forall n, (even n -> n = 2*(div2 n)) /\ (odd n -> n = 2*(div2 n) +1).
  Proof.
    apply nat_2_ind; split; intros H0; simpl; auto.
    contradiction (not_even_odd 0); auto.
    contradiction (not_even_odd 1); auto.
    destruct H as [IH1 IH2].
    inversion_clear H0. 
    inversion_clear H. 
    rewrite (IH1 H0) at 1. simpl. omega.
    destruct H as [IH1 IH2].
    inversion_clear H0.
    inversion_clear H.
    rewrite (IH2 H0) at 1. simpl. omega.
  Qed.

  Theorem div2_thm_even n : even n -> n = 2*(div2 n).
  Proof. apply div2_lemma_rec. Qed.

  Theorem div2_thm_odd n : odd n -> n = 2*(div2 n)+1.
  Proof. apply div2_lemma_rec. Qed.

  Theorem div2_inf n : 0 < n -> div2 n < n.
  Proof.
    intros H0.
    destruct (even_or_odd n) as [ H1 | H1 ].
    apply div2_thm_even in H1.
    rewrite H1 at 2.
    generalize H1. clear H1.
    set (x := div2 n).
    intros H1.
    destruct (le_lt_or_eq 0 x); omega.
    apply div2_thm_odd in H1.
    rewrite H1 at 2. omega.
  Qed.

End div2_even.

Section Power2.

  Fixpoint power2 n := match n with
    | 0 => 1
    | S n => 2*(power2 n)
  end.

  Proposition power2_inc n : n < power2 n.
  Proof.
    induction n; simpl. auto.
    omega.
  Qed.

  Proposition power2_0 n : 0 < power2 n.
  Proof. induction n; simpl; omega. Qed.

End Power2.

Section CompDecomp.
   
  Let comp_uniq_rec b b' : forall a a', a < a'  -> (power2 a)*b = (power2 a')*b' -> odd b -> odd b' -> False.
  Proof.
    induction a; intros a' H0 H1 Hb Hb'.
    simpl in H1; rewrite plus_comm in H1; simpl in H1.
    generalize H0 H1. elim a'. intros C. contradict C. omega.
    intros n _ _ Hn. unfold power2 in Hn. fold power2 in Hn.
    rewrite Hn in Hb. apply not_even_odd with (2 := Hb).
    rewrite <- mult_assoc.
    apply even_2x.
    generalize H0 H1. clear H0 H1. elim a'.
    intros H. contradict H. omega.
    clear a'. intros a' _ Ha' H1.
    unfold power2 in H1; fold power2 in H1.
    rewrite <- mult_assoc in H1.
    rewrite <- mult_assoc in H1.
    apply lt_S_n in Ha'.
    apply (IHa  _ Ha'); auto.
    apply mult_reg_l in H1; auto.
  Qed.

  Let comp_uniq_bis a a' b b' : odd b -> odd b' -> (power2 a)*b = (power2 a')*b' -> a = a'.
  Proof.
    intros Hb Hb' Hp.
    destruct (lt_eq_lt_dec a a') as [ [ H | H ] | H ]; auto; [ | symmetry in Hp ];
    exfalso; apply (comp_uniq_rec _ _ _ _ H Hp); auto.
  Qed.
 
  Theorem comp_uniq a a' b b' : (power2 a)*(2*b+1) = (power2 a')*(2*b'+1) -> (a=a' /\ b=b').
  Proof.
    intros H.
    assert (a=a') as Ha.
    apply comp_uniq_bis with (3 := H); apply odd_2x1.
    split; auto.
    rewrite <- Ha in H.
    apply mult_reg_l in H.
    omega.
    apply power2_0.
  Qed.

  Proposition comp_lt_a a b : a < (power2 a)*(2*b+1).
  Proof.
    unfold lt.
    assert (S a = (S a)*1) as H. 
    rewrite mult_comm; simpl; auto. 
    rewrite H at 1; clear H.
    apply mult_le_compat.
    apply power2_inc.
    omega.
  Qed.

  Proposition comp_lt_b a b : b < (power2 a)*(2*b+1).
  Proof.
    unfold lt.
    assert (S b = 1*(S b)) as H. 
    simpl; auto. 
    rewrite H at 1; clear H.
    apply mult_le_compat.
    apply power2_0.
    omega.
  Qed.

  Proposition comp_lt_0 a b : 0 < (power2 a)*(2*b+1).
  Proof.
    apply le_lt_trans with (1 := le_0_n a).
    apply comp_lt_a.
  Qed.
 
  Definition decomp : forall n, 0 < n -> { a:nat  & { k:nat | n = (power2 a)*(2*k+1) } }.
  Proof.
    apply (nat_gen_rec (fun n => 0 < n -> { a:nat & { k | n = (power2 a)*(2*k+1) } } )).
    intros n IH Hn0.
    destruct (even_or_odd n) as [ Hn1 | Hn1 ].
    apply div2_thm_even in Hn1.
    destruct (IH (div2 n) (div2_inf _ Hn0)) as [ a Ha ]. omega.
    destruct Ha as [ k Hk ].
    exists (S a). exists k. unfold power2. fold power2. 
    rewrite Hn1, Hk, mult_assoc. omega.
    apply div2_thm_odd in Hn1.
    exists 0. exists (div2 n). simpl. omega.
  Qed.

  Definition decomp_a n (H : 0 < n) : nat := projT1 (decomp _ H).
  Definition decomp_b n (H : 0 < n) : nat := proj1_sig (projT2 (decomp _ H)).

  Theorem decomp_ab_prop n (H : 0 < n) : n = (power2 (decomp_a _ H))*(2*(decomp_b _ H)+1).
  Proof.
    unfold decomp_a, decomp_b.
    apply (proj2_sig (projT2 (decomp _ H))).
  Qed.

End CompDecomp.

Section Nat2.

  Let Sinj x y : S x = S y -> x = y.
  Proof. omega. Qed.

  Let f n := let Pn := lt_0_Sn n in (decomp_a _ Pn, decomp_b _ Pn).
 
  Let g c := match c with (a,b) => (power2 a)*(2*b+1) -1 end.

  Let pred_inj a b : 0 < a -> 0 < b -> a-1 = b-1 -> a = b.
  Proof. omega. Qed.

  Let g_inj : forall x y, g x = g y -> x = y.
  Proof.
    intros (a,b) (a',b').
    unfold g. intros H.
    apply pred_inj in H; (apply comp_lt_0 || trivial).
    apply comp_uniq in H.
    destruct H as [ H1 H2 ]. 
    rewrite H1, H2; trivial.
  Qed.

  Let gf x : g (f x) = x.
  Proof.
    unfold f, g.
    rewrite <- decomp_ab_prop. omega.
  Qed.

  Let fg x : f (g x) = x.
  Proof.
    apply g_inj.
    rewrite gf.
    trivial.
  Qed.

  Definition nat2_index : indexation (nat * nat).
  Proof. exists f g; trivial. Qed.

End Nat2.






  
    
    
    