(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.

Require Import maps.

Section Defs.

  Variable f : nat -> nat.
  
  Definition fnat_sinc := forall x y, x < y -> f x < f y.
  Definition fnat_inc  := forall x y, x <= y -> f x <= f y.
  Definition fnat_inj  := injective f.

  Theorem fnat_inc_imp_sinc : (forall x, f x < f (S x)) -> fnat_sinc.
  Proof.
    intros H. red. 
    induction y; intros Hx.
    contradict Hx; auto with arith.
    destruct (le_lt_dec y x) as [ Hy | Hy ].
    cutrewrite (x = y); auto with arith.
    apply le_S_n in Hx; auto with arith.
    apply lt_trans with (1 := IHy Hy); auto.
  Qed.

  Hypothesis f_sinc : fnat_sinc.

  Proposition fnat_sinc_imp_inc : fnat_inc.
  Proof.
    intros x y Hxy.
    destruct (le_lt_dec y x) as [ H | H ].
    cutrewrite (x=y); auto with arith.
    apply f_sinc in H; auto with arith.
  Qed.

  Proposition fnat_sinc_rev : forall x y, f x < f y -> x < y.
  Proof.
    intros x y Hxy.
    destruct (le_lt_dec y x) as [ H | H ]; auto.
    apply fnat_sinc_imp_inc in H.
    contradict H; auto with arith.
  Qed.

  Proposition fnat_sinc_imp_inj : fnat_inj.
  Proof.
    intros x y Hxy.
    destruct (lt_eq_lt_dec x y) as [ [ H | H ] | H ]; trivial;
    apply f_sinc in H; contradict H; rewrite Hxy; auto with arith.
  Qed.
  
  Proposition fnat_inc_rev : forall x y, f x <= f y -> x <= y.
  Proof.
    intros x y Hxy.
    destruct (le_lt_dec x y) as [ H | H ]; trivial.
    apply f_sinc in H; contradict H; auto with arith.
  Qed.

  Definition fnat_find n : f 0 <= n -> { x | f x <= n /\ n < f (S x) }.
  Proof.
    induction n; intros Hn.
    exists 0; split; trivial.
    assert (0 = f 0) as H; auto with arith.
    rewrite H at 1; auto.
    destruct (le_lt_dec (S n) (f 0)) as [H | H].
    exists 0. rewrite (le_antisym _ _ H Hn).
    split; auto.
    destruct IHn as [x [ H1 H2 ]].
    apply le_S_n in H; auto.
    destruct (le_lt_dec (f (S x)) (S n)).
    exists (S x); split; trivial. 
    apply le_lt_trans with (1 := H2). auto. 
    exists x; split; auto with arith.
  Qed.

  Definition fnat_sinc_inv n := match le_lt_dec (f 0) n with
    | left  H => proj1_sig (fnat_find _ H)
    | right _ => 0
  end.

  Proposition fnat_sinc_inv_prop : forall x, fnat_sinc_inv (f x) = x.
  Proof.
    intros x.
    unfold fnat_sinc_inv.
    destruct (le_lt_dec (f 0) (f x)) as [H | H].
    generalize (proj2_sig (fnat_find (f x) H)).
    set (y := proj1_sig (fnat_find (f x) H)).
    intros [ H1 H2 ].
    apply fnat_inc_rev in H1.
    apply fnat_sinc_rev in H2.
    apply le_S_n in H2; auto with arith.
    apply fnat_sinc_rev in H.
    contradict H; auto with arith.
  Qed.

End Defs.


