(**************************************************************)
(*   The Strong Completeness of Partial Monoidal BI           *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.

Require Import list_fall_exst.
Require Import list_in_t.

Require Import predicate.
Require Import closure.
Require Import maps.
Require Import carrier.
Require Import finite.

Require Import system.
Require Import system_extensions.

Require Import words.
Require Import constraints.

Section PMO.

  Set Implicit Arguments.

  Variable letter : Type.

  Section Constraint_rules.

    Section rules_prems_concs.

      Definition c_rule_eps_c := (eps ~~ eps : constraint letter).

      Variable p : (word letter*word letter*word letter)%type.

      Let x := match p with (x,_,_) => x end.
      Let y := match p with (_,y,_) => y end.
      Let z := match p with (_,_,z) => z end.
      
      Definition c_rule_comm_p1 := x o y ~~ x o y.
      Definition c_rule_comm_c  := x o y ~~ y o x.
      
      Definition c_rule_lft_p1  := x ~~ y.
      Definition c_rule_lft_c   := x ~~ x.
      
      Definition c_rule_rt_p1  := x ~~ y.
      Definition c_rule_rt_c   := y ~~ y.
      
      Definition c_rule_tran_p1 := x ~~ y.
      Definition c_rule_tran_p2 := y ~~ z.
      Definition c_rule_tran_c  := x ~~ z.
      
      Definition c_rule_sub_p1  := x o y ~~ x o y.
      Definition c_rule_sub_c   :=   x   ~~   x.
      
      Definition c_rule_comp_p1 := z o y ~~ z o y.
      Definition c_rule_comp_p2 :=   x   ~~   y.
      Definition c_rule_comp_c  := z o x ~~ z o y.

    End rules_prems_concs.
    
    Definition c_rule_eps  := MkUAxiom  c_rule_eps_c.
    Definition c_rule_comm := MkUSingle c_rule_comm_p1 c_rule_comm_c.
    Definition c_rule_lft  := MkUSingle c_rule_lft_p1 c_rule_lft_c.
    Definition c_rule_rt   := MkUSingle c_rule_rt_p1 c_rule_rt_c.
    Definition c_rule_tran := MkUDouble c_rule_tran_p1 c_rule_tran_p2 c_rule_tran_c.
    Definition c_rule_sub  := MkUSingle c_rule_sub_p1 c_rule_sub_c.
    Definition c_rule_comp := MkUDouble c_rule_comp_p1 c_rule_comp_p2 c_rule_comp_c.
    
    Definition c_sys_PMO := c_rule_eps::c_rule_comm::c_rule_lft::c_rule_rt::c_rule_tran::c_rule_sub::c_rule_comp::nil.
    
    Definition is_pmo := closed_under_rules  c_sys_PMO.
    Definition pmo    := closure_under_proof c_sys_PMO.

    (* these are generic properties that hold for ANY list of rules *)
    
    Proposition pmo_is_pmo R : is_pmo (pmo R).
    Proof. apply closure_under_proof_is_closed. Qed.
    
    Proposition is_pmo_postfixpoint R : is_pmo R <-> pmo R inc R.
    Proof.
      split.
      apply closed_under_rules_postfixpoint.
      apply postfixpoint_closed_under_rules.
    Qed.
    
    Proposition pmo_mono R S : R inc S -> pmo R inc pmo S.
    Proof. apply closure_under_proof_is_mono. Qed.

    Proposition pmo_inc R : R inc pmo R.
    Proof. apply closure_under_proof_is_inc. Qed.
    
    Proposition pmo_idem R : pmo (pmo R) inc pmo R. 
    Proof. apply closure_under_proof_is_idem. Qed.

    Proposition pmo_mono' : Closure_monotonic pmo.        Proof. exact pmo_mono. Qed.
    Proposition pmo_inc'  : Closure_increasing pmo.       Proof. exact pmo_inc. Qed.
    Proposition pmo_idem' : Closure_idempotent pmo.       Proof. exact pmo_idem. Qed.
    
    Proposition pmo_cpct : Closure_compact pmo.
    Proof. intros C x H; destruct closure_under_proof_compactness with (1 := H) as ( Rf ); exists Rf; tauto. Qed.
    
    Proposition pmo_min C seq : is_pmo seq -> C inc seq -> pmo C inc seq.
    Proof.
      intros H1 ?; apply is_pmo_postfixpoint in H1.
      apply inc_transitive with (2 := H1), pmo_mono; auto.
    Qed.  

    Proposition pmo_cl_inc R S : R inc pmo S -> pmo R inc pmo S.
    Proof. apply pmo_min, pmo_is_pmo. Qed.

  End Constraint_rules.  

  Hint Resolve pmo_is_pmo is_pmo_postfixpoint pmo_mono pmo_inc pmo_idem pmo_cpct pmo_min.

  Proposition is_pmo_fixpoint R : is_pmo R -> R ~ pmo R.
  Proof. split; auto. Qed.

  Proposition pmo_compact Cf R : finite Cf -> Cf inc pmo R -> exists Rf, finite Rf /\ Rf inc R /\ Cf inc pmo Rf.
  Proof. apply cl_compact; auto. red; apply pmo_mono. Qed.

  Proposition pmo_id R S : R ~ S -> pmo R ~ pmo S.
  Proof. intros [ ]; split; auto. Qed.
  
  Proposition is_pmo_id R S : R ~ S -> is_pmo R -> is_pmo S.
  Proof.
    intros [ H1 H2 ] H3.
    apply is_pmo_fixpoint in H3.
    apply is_pmo_postfixpoint.
    apply inc_transitive with (2 := H1),
          inc_transitive with (2 := proj2 H3); auto.
  Qed.
  
  Hint Resolve pmo_id is_pmo_id.
    
  Section is_pmo_closed.

    Variable ceq : constraint letter -> Prop.
    Hypothesis ceqPMO : is_pmo ceq.
  
    Ltac solve r p := 
      intros;
      match goal with [ _: ceq ?x, _: ceq ?y |- _ ] => apply (@ceqPMO r (x::y::nil))
                    | [ _: ceq ?x            |- _ ] => apply (@ceqPMO r (x::nil))
                    | [                      |- _ ] => apply (@ceqPMO r nil)
      end; unfold c_sys_PMO; In_t_auto; exists p; simpl; auto.

    Proposition ceq_eps : ceq (eps ~~ eps).
    Proof. solve c_rule_eps tt. Qed.
    
    Proposition ceq_comm x y : ceq (x o y ~~ x o y) -> ceq (x o y ~~ y o x).
    Proof. solve c_rule_comm (x,y,y). Qed.
    
    Proposition ceq_sub x y : ceq (x o y ~~ x o y) -> ceq (x ~~ x).
    Proof. solve c_rule_sub (x,y,y). Qed.
    
    Proposition ceq_comp k x y : ceq (k o y ~~ k o y) -> ceq (x ~~ y) -> ceq (k o x ~~ k o y).
    Proof. solve c_rule_comp (x,y,k). Qed.
    
    Proposition ceq_tran x y z : ceq (x ~~ y) -> ceq (y ~~ z) -> ceq (x ~~ z).
    Proof. solve c_rule_tran (x,y,z). Qed.
    
    Proposition ceq_lft x y : ceq (x ~~ y) -> ceq (x ~~ x).
    Proof. solve c_rule_lft (x,y,y). Qed.

    Proposition ceq_rt x y : ceq (x ~~ y) -> ceq (y ~~ y).
    Proof. solve c_rule_rt (x,y,y). Qed.
    
    (* hence pmo are closed under the intented rules, as expected *)
    
    Hint Resolve ceq_eps ceq_comm ceq_sub ceq_lft ceq_rt.
    
    Proposition ceq_comm_l x y m : ceq ( x o y ~~ m ) -> ceq (y o x ~~ m).
    Proof. 
      intros H.
      generalize (ceq_lft H); intros H'.
      apply ceq_comm, ceq_rt, ceq_comm in H'.
      apply ceq_tran with (1 := H'); auto. 
    Qed.
    
    Proposition ceq_comm_r x y m : ceq ( m ~~ x o y ) -> ceq (m ~~ y o x).
    Proof. 
      intros H.
      generalize (ceq_rt H); intros H'.
      apply ceq_comm in H'.
      apply ceq_tran with (2 := H'); auto. 
    Qed.

    Proposition ceq_comm' x y : ceq (x o y ~~ x o y) -> ceq (y o x ~~ x o y).
    Proof. intros; apply ceq_comm_l; auto. Qed.

    Proposition ceq_pl k x y: ceq (k o x ~~ y) -> ceq (x ~~ x).
    Proof. intros H; apply ceq_lft, ceq_comm, ceq_rt, ceq_sub in H; auto. Qed.
    
    Proposition ceq_pr k x y : ceq (x ~~ k o y) -> ceq (y ~~ y).
    Proof. intros H; apply ceq_comm_r, ceq_rt, ceq_sub in H; auto. Qed.
    
    Hint Resolve ceq_pl ceq_pr.
    
    Proposition ceq_el k x y m : ceq (x ~~ y) -> ceq (k o y ~~ m) -> ceq (k o x ~~ m).
    Proof.
      intros H1 H2; generalize (ceq_lft H2); intros H3.
      apply ceq_tran with (2 := H2), ceq_comp; auto.
    Qed.
    Proposition ceq_el' k x y m : ceq (x ~~ y) -> ceq (y o k ~~ m) -> ceq (x o k ~~ m).
    Proof.
      intros H1 H2.
      apply ceq_comm_l in H2.
      apply ceq_comm_l, ceq_el with (2 := H2); auto.
    Qed.

    Let ceq_weq_ech_l m k x y : ceq ((x o y) o k ~~ m) -> ceq ((y o x) o k ~~ m).
    Proof.
      intros H.
      apply ceq_comm_l.
      apply ceq_comm_l in H.
      apply ceq_tran with (2 := H).
      apply ceq_comp.
      apply ceq_lft with (1 := H).
      apply ceq_comm'.
      apply ceq_pl with (1 := H).
    Qed.
 
    Let ceq_weq_ech_r m k x y : ceq (m ~~ (x o y) o k) -> ceq (m ~~ (y o x) o k).
    Proof.
      intros H.
      apply ceq_comm_r.
      apply ceq_tran with (y := k o (x o y)); auto.
      apply ceq_comm_r; auto.
      apply ceq_rt, ceq_weq_ech_l, ceq_comm_l in H.
      apply ceq_comm_l, ceq_weq_ech_l, ceq_comm_l, ceq_lft with (1 := H).
    Qed.

    Proposition ceq_weq_l x y m : x weq y -> ceq (x ~~ m) -> ceq (y ~~ m).
    Proof.
      intros H; generalize H m; clear m H.
      induction 1 as [ | a x x' | a b x | ]; intros m; auto.
      cutrewrite (a::x = l a o x); auto.
      cutrewrite (a::x' = l a o x'); auto.
      intros H1.
      generalize (IHPermutation _ (ceq_pl _ _ H1)); intros H2.
      apply ceq_el with (1 := H2); auto.
      intros H1.
      generalize (ceq_lft H1).
      intros H2.
      cutrewrite (a::b::x = (l a o l b) o x); auto.
    Qed.
 
    Let ceq_weq_r_rec x y : x weq y -> forall m k, ceq (m ~~ k o x) -> ceq (m ~~ k o y).
    Proof.
      induction 1 as [ | a x x' | a b x | ]; intros m k; auto.
      cutrewrite (k++a::x = (k o l a) o x).
      cutrewrite (k++a::x' = (k o l a) o x').
      apply IHPermutation.
      rewrite app_ass; simpl; auto.
      rewrite app_ass; simpl; auto.
      intros H2.
      cutrewrite (a::b::x = (l a o l b) o x); auto.
      cutrewrite (b::a::x = (l b o l a) o x) in H2; auto.
      apply ceq_comm_r.
      rewrite app_ass.
      apply ceq_weq_ech_r.
      rewrite <- app_ass.
      apply ceq_comm_r; auto.
    Qed.
   
    Proposition ceq_weq_r x y m : x weq y -> ceq (m ~~ x) -> ceq (m ~~ y).
    Proof. intros H; apply ceq_weq_r_rec with (k := eps) (1 := H). Qed.
    
    Hint Resolve ceq_weq_l ceq_weq_r.
    
    Proposition ceq_weq x1 x2 y1 y2 : x1 weq x2 -> y1 weq y2 -> ceq (x1 ~~ y1) -> ceq (x2 ~~ y2).
    Proof. intros; eauto. Qed.
    
    Proposition ceq_language_1 x : language ceq x -> ceq (x ~~ x).
    Proof.
      intros Hx.
      destruct Hx as ( a & b & Hab & [ [ k Hx ] | [ k Hx ] ] ).
      apply weq_sym in Hx.
      apply ceq_weq_l with (1 := Hx) in Hab.
      apply ceq_comm_l in Hab.
      apply ceq_pl in Hab; auto.
      apply weq_sym in Hx.
      apply ceq_weq_r with (1 := Hx) in Hab.
      apply ceq_comm_r in Hab.
      apply ceq_pr in Hab; auto.
    Qed.
    
    Proposition ceq_language_2 x : ceq (x ~~ x) -> language ceq x.
    Proof.
      intros Hx.
      exists x; exists x.
      split; auto.
      left; exists eps. 
      rewrite -> eps_neutral_r; auto.
    Qed.
  
    Proposition ceq_alphabet_1 a : alphabet ceq a -> ceq (l a ~~ l a).
    Proof. apply ceq_language_1. Qed.
    
    Proposition ceq_alphabet_2 a : ceq (l a ~~ l a) -> alphabet ceq a.
    Proof. apply ceq_language_2. Qed.
    
  End is_pmo_closed.
  
  Section closed_is_pmo.

  (* every relation closed under the following rules is a pmo *)  
  
    Variable meq : constraint letter -> Prop.

    Hypothesis meq_eps  : meq (eps ~~ eps).
    Hypothesis meq_comm : forall x y, meq (x o y ~~ x o y) -> meq (x o y ~~ y o x).
    Hypothesis meq_sub  : forall x y, meq (x o y ~~ x o y) -> meq (x ~~ x).
    Hypothesis meq_comp : forall k x y, meq (k o y ~~ k o y) -> meq (x ~~ y) -> meq (k o x ~~ k o y).
    Hypothesis meq_tran : forall x y z, meq (x ~~ y) -> meq (y ~~ z) -> meq (x ~~ z).
    Hypothesis meq_lft  : forall x y, meq (x ~~ y) -> meq (x ~~ x).
    Hypothesis meq_rt   : forall x y, meq (x ~~ y) -> meq (y ~~ y). 
    
    Proposition meq_is_pmo : is_pmo meq.
    Proof.
      intros r hh c Hr I HH.
      unfold c_sys_PMO in Hr; In_t_auto; rewrite Hr in I; red in I; simpl in I;
      destruct I as [ x [ _ H ] ]; 
      injection  H; intros H1 H2; rewrite H1; rewrite H2 in HH; clear H1 H2 H;
      (try apply meq_eps); destruct x as [ [ x y ] z ]; list_fall_auto;
      repeat match goal with [ K: meq (?cr _) |- _ ] => unfold cr in K 
                            |[ |- meq (?cr _)      ] => unfold cr end; auto.
      apply meq_lft with (1 := H).
      apply meq_rt with (1 := H).
      apply meq_tran with (y := y); auto.
      apply meq_sub with (y := y); auto.
    Qed.
  
  End closed_is_pmo.

  Section pmo_props.
  
    Ltac solve x := apply x with (1 := @pmo_is_pmo _).
  
    Proposition pmo_eps R         : pmo R (eps ~~ eps).      Proof. solve ceq_eps. Qed.
    
    Proposition pmo_comm R x y    : pmo R (x o y ~~ x o y) 
                                 -> pmo R (x o y ~~ y o x).  Proof. solve ceq_comm. Qed.

    Proposition pmo_comm' R x y   : pmo R (x o y ~~ x o y) 
                                 -> pmo R (y o x ~~ x o y).  Proof. solve ceq_comm'. Qed.
                                 
    Proposition pmo_sub R x y     : pmo R (x o y ~~ x o y) 
                                 -> pmo R     (x ~~ x).      Proof. solve ceq_sub. Qed.

    Proposition pmo_comp R k x y  : pmo R (k o y ~~ k o y) 
                                 -> pmo R     (x ~~ y)
                                 -> pmo R (k o x ~~ k o y).  Proof. solve ceq_comp. Qed.
 
    Proposition pmo_tran R x y z  : pmo R (x ~~ y)
                                 -> pmo R (y ~~ z) 
                                 -> pmo R (x ~~ z).          Proof. solve ceq_tran. Qed.

    Proposition pmo_lft R x y     : pmo R (x ~~ y) 
                                 -> pmo R (x ~~ x).          Proof. solve ceq_lft. Qed.

    Proposition pmo_rt R x y      : pmo R (x ~~ y) 
                                 -> pmo R (y ~~ y).          Proof. solve ceq_rt. Qed.

    Proposition pmo_pl R k x y    : pmo R (k o x ~~ y) 
                                 -> pmo R     (x ~~ x).      Proof. solve ceq_pl. Qed.

    Proposition pmo_pr R k x y    : pmo R (x ~~ k o y) 
                                 -> pmo R (y ~~ y).          Proof. solve ceq_pr. Qed.

    Proposition pmo_el R k x y m  : pmo R (x ~~ y) 
                                 -> pmo R (k o y ~~ m) 
                                 -> pmo R (k o x ~~ m).      Proof. solve ceq_el. Qed.
      
    Proposition pmo_el' R k x y m : pmo R (x ~~ y) 
                                 -> pmo R (y o k ~~ m) 
                                 -> pmo R (x o k ~~ m).      Proof. solve ceq_el'. Qed.

    Proposition pmo_weq R x1 x2 y1 y2 : x1 weq x2 -> y1 weq y2 
                                     -> pmo R (x1 ~~ y1) 
                                     -> pmo R (x2 ~~ y2).    Proof. solve ceq_weq. Qed.

    Proposition language_pmo C x : language (pmo C) x <-> pmo C (x ~~ x).
    Proof. split. solve ceq_language_1. apply ceq_language_2. Qed.
    
    Proposition pmo_language C x : pmo C (x ~~ x) -> language (pmo C) x.
    Proof. apply ceq_language_2; auto. Qed.
    
    Proposition pmo_alphabet C a : alphabet (pmo C) a <-> pmo C (l a ~~ l a).
    Proof. split. solve ceq_alphabet_1. apply ceq_alphabet_2. Qed.
    
  End pmo_props.
  
  Hint Resolve pmo_eps pmo_comm pmo_comm' pmo_comp.
  Hint Resolve pmo_el pmo_el' pmo_weq.
  Hint Resolve language_pmo pmo_language pmo_alphabet.
  
  Proposition pmo_eps_1 R : pmo (R cup sg (eps~~eps)) ~ pmo R.
  Proof. split; auto. Qed.
  
  Hint Resolve pmo_eps_1.

  Section pmo_help.

    Variable C : constraint letter -> Prop.
    Variables x y z : word letter.
    
    Hypothesis HC : pmo C (x o y ~~ z).
    
    Fact pmo_help_lemma_1 : pmo C (y ~~ y).
    Proof. apply pmo_pl in HC; auto. Qed.
    
    Fact pmo_help_lemma_2 : pmo C (x ~~ x).
    Proof. apply pmo_lft, pmo_comm', pmo_pl in HC; auto. Qed.
  
    Fact pmo_help_lemma_4 : pmo C (z ~~ z).
    Proof. apply pmo_rt in HC; auto. Qed.

    Hint Resolve pmo_help_lemma_1 pmo_help_lemma_2 pmo_help_lemma_4.
    
    Fact pmo_help_lemma_9 : pmo C (x ~~ x) /\ pmo C (y ~~ y) /\ pmo C (z ~~ z).
    Proof. auto. Qed.

  End pmo_help.

  Fact pmo_help_lemma_3 C x y m : pmo C (x o m ~~ y) -> pmo C (x ~~ x).
  Proof. apply pmo_help_lemma_2. Qed.
  
  Fact pmo_help_lemma_5 C1 C2 c : pmo C2 c -> pmo (C1 cup C2) c.
  Proof. apply pmo_mono; auto. Qed.
  
  Fact pmo_help_lemma_6 a m k : pmo (sg ((a::m) ~~ k)) (l a ~~ l a).
  Proof.
    cutrewrite (a::m = l a o m); auto.
    apply pmo_help_lemma_2 with (y := m) (z := k), pmo_inc; auto.
  Qed.
  
  Fact pmo_help_lemma_7 a m k : pmo (sg ((a::m) ~~ k)) (m ~~ m).
  Proof.
    cutrewrite (a::m = l a o m); auto.
    apply pmo_help_lemma_1 with (x := l a) (z := k), pmo_inc; auto.
  Qed.

  Fact pmo_help_lemma_8 m k : pmo (sg (m ~~ k)) (k ~~ k).
  Proof. apply pmo_rt with (x := m), pmo_inc; auto. Qed.

  Section meq_L.
  
    Variable L : letter -> Prop.
  
    Definition meq_L (c : constraint letter) := match c with (m,n) => carrier m inc L /\ carrier n inc L end.
    
    Proposition R_inc_meq_L R : alphabet R inc L -> R inc meq_L.
    Proof.
      intros HR (x,y) Hxy.
      split; apply inc_transitive with (B := alphabet R); auto;
      intros a Ha; apply carrier_inc_subletter in Ha;
      exists x; exists y; tauto.
    Qed.
    
    Proposition meq_L_is_pmo : is_pmo meq_L.
    Proof.
      apply meq_is_pmo; simpl; intros; pred_solve; 
      try (apply carrier_app_1 in H; destruct H; auto).
    Qed.
     
     Proposition meq_L_postfixpoint : pmo meq_L inc meq_L.
     Proof. apply is_pmo_postfixpoint, meq_L_is_pmo. Qed.
     
     Proposition alphabet_meq_L_1 : alphabet meq_L inc L.
     Proof. intros ? (?,(?,([ ],[ Ha | Ha ]))); apply subletter_inc_carrier in Ha; auto. Qed.
     
     Proposition alphabet_meq_L_2 : L inc alphabet meq_L.
     Proof.
       intros a Ha. 
       assert (carrier (a::nil) inc L); auto.
       exists (l a); exists (l a). 
       split.
       simpl; auto.
       left; exists eps; simpl; auto.
     Qed.
     
     Hint Resolve alphabet_meq_L_1 alphabet_meq_L_2 .
     
     Proposition alphabet_meq_L : L ~ alphabet meq_L.
     Proof. split; auto. Qed.
     
     Proposition meq_L_o m n k : meq_L (m~~m) -> meq_L (n~~n) -> meq_L (k~~k) -> meq_L (m o n ~~ k).
     Proof. intros [] [] []; split; auto. Qed.
     
     Proposition meq_L_eps_eps : meq_L (eps~~eps).
     Proof. split; auto. Qed.
     
     Proposition meq_L_eps m : meq_L (m~~m) -> meq_L (eps~~m).
     Proof. intros []; split; auto. Qed.

  End meq_L.
  
  Hint Resolve R_inc_meq_L alphabet_meq_L alphabet_meq_L_1 alphabet_meq_L_2.
  
  Proposition meq_L_inc L L' : L inc L' -> meq_L L inc meq_L L'.
  Proof. intros ? [] []; split; apply inc_transitive with (B := L); auto. Qed.

  Hint Resolve meq_L_inc.
  
  Proposition alphabet_inc_meq_L R L : alphabet R inc L -> pmo R inc meq_L L.
  Proof. intros; apply inc_transitive with (2 := proj1 (is_pmo_postfixpoint _) (@meq_L_is_pmo _)); auto. Qed.
  
  Hint Resolve alphabet_inc_meq_L.
  
  Proposition pmo_inc_alphabet R m : pmo R (m~~m) -> carrier m inc alphabet R.
  Proof.
    intros Hm.
    apply alphabet_inc_meq_L with (L := alphabet R) in Hm; auto.
    apply Hm.
  Qed. 
  
  Proposition alphabet_pmo R : alphabet R ~ alphabet (pmo R).
  Proof. split; auto; apply inc_transitive with (B := alphabet (meq_L (alphabet R))); auto. Qed.
  
  Proposition pmo_empty : pmo empty ~ sg (eps~~eps).
  Proof.
    split.
    intros (x,y) Hxy.
    apply (@alphabet_inc_meq_L empty empty) in Hxy.
    destruct Hxy as [ H0 H1 ].
    apply carrier_eps in H0. 
    apply carrier_eps in H1.
    rewrite H0, H1; auto.
    apply alphabet_empty.
    pred_solve.
  Qed.
  
  Proposition alphabet_pmo_finite R : finite R -> finite (alphabet (pmo R)).
  Proof. intro; apply finite_id with (1 := alphabet_pmo _); auto. Qed.

End PMO.

Hint Resolve pmo_is_pmo is_pmo_postfixpoint.
Hint Resolve pmo_mono pmo_inc pmo_idem.
Hint Resolve pmo_mono' pmo_inc' pmo_idem'.
Hint Resolve pmo_cpct pmo_min pmo_cl_inc.

Hint Extern 0 ( ?x inc pmo ?x ) => apply pmo_inc.
Hint Extern 1 ( pmo _ inc pmo _ ) => apply pmo_mono.
Hint Extern 1 ( pmo _ inc pmo _ ) => apply pmo_cl_inc. 
Hint Extern 1 ( _ inc pmo _ ) => apply inc_transitive with (2 := pmo_inc _).
Hint Extern 1 (pmo _ _) => apply pmo_inc.
	
Hint Resolve pmo_id.
Hint Resolve pmo_eps pmo_comm pmo_comm' pmo_comp.
Hint Resolve language_pmo pmo_language pmo_alphabet.
Hint Resolve pmo_eps_1.
Hint Resolve R_inc_meq_L alphabet_meq_L alphabet_meq_L_1 alphabet_meq_L_2.
Hint Resolve meq_L_inc.
Hint Resolve alphabet_inc_meq_L.
Hint Resolve pmo_inc_alphabet alphabet_pmo alphabet_pmo_finite.
  
Section Substitutions.

  Set Implicit Arguments.

  Variable L M : Type.  

  Variable sigma : L -> word M.
  Let s := subst_w sigma.
  
  Definition subst_cst (c : constraint L) := match c with (m,n) => s m ~~ s n end.   
  Definition subst_rimage (C : _ -> Prop) c := match c with (m,n) => C (s m ~~ s n) end.
  Definition subst_dimage := pd_image subst_cst.
  
  Proposition subst_cst_rimage C : subst_rimage C ~ pr_image subst_cst C.
  Proof.
    split; intros (m,n) Hmn.
    exists (s m ~~ s n); tauto.
    destruct Hmn as [ (x,y) (H1,H2) ].
    rewrite <- H2 in H1; auto.
  Qed.
  
  Proposition subst_dimage_inc C C' : C inc C' -> subst_dimage C inc subst_dimage C'.
  Proof. intros ?; apply pd_image_inc; auto. Qed.
  
  Proposition subst_rimage_inc C C' : C inc C' -> subst_rimage C inc subst_rimage C'.
  Proof. 
    intros ?. 
    apply inc_transitive with (1 := proj1 (subst_cst_rimage _)).
    apply inc_transitive with (2 := proj2 (subst_cst_rimage _)).
    apply pr_image_inc; auto. 
  Qed.
  
  Proposition subst_rimage_is_pmo C : is_pmo C -> is_pmo (subst_rimage C).
  Proof.
    intros HC.
    apply meq_is_pmo; unfold subst_rimage.
    
    apply ceq_eps with (1 := HC).
    
    intros x y H; unfold s in H |- *. 
    rewrite subst_o in H.
    rewrite subst_o, subst_o. 
    apply ceq_comm with (1 := HC); auto.
    
    intros x y H; unfold s in H |- *. 
    rewrite subst_o in H.
    apply ceq_sub with (1 := HC) (2 := H); auto.
    
    intros k x y H1 H2; unfold s in H1 |- *. 
    rewrite subst_o in H1.
    rewrite subst_o, subst_o.
    apply ceq_comp with (1 := HC); auto.
    
    intros ? ? ?; apply ceq_tran; auto.
    
    intros ? ?; apply ceq_lft; auto.
    
    intros ? ?; apply ceq_rt; auto.
  Qed.
  
  Variable C : constraint L -> Prop.
  
  Proposition subst_dimage_alphabet : alphabet (subst_dimage C) ~ subst_alphabet sigma (alphabet C).
  Proof.
    split.

    intros a ( x & y & [ (m,n) [ Hmn Hxy ] ] & Ha ).
    injection Hxy; intros Hn Hm; rewrite <- Hm, <- Hn in * |- *; clear x y Hxy Hm Hn.
    destruct Ha as [ Ha | Ha ]; apply subletter_inc_carrier in Ha; [ exists m | exists n ]; split; eauto.

    intros a Ha; apply subst_alphabet_prop in Ha. 
    destruct Ha as ( b & ( m & n & Hmn & [ Hb | Hb ] ) & Ha ); apply subletter_inc_carrier in Hb;
    exists (s m); exists (s n); split; 
    try ( exists (m,n); tauto ); [ left | right ];
    apply carrier_inc_subletter, subst_carrier with (2 := Ha); auto. 
  Qed.
  
  Proposition subst_dimage_pmo_inc : subst_dimage (pmo C) inc pmo (subst_dimage C).
  Proof.
    set (R := subst_rimage (pmo (subst_dimage C))).
    assert (is_pmo R) as H0.  
      apply subst_rimage_is_pmo; auto.
    assert (C inc subst_rimage (pmo (subst_dimage C))) as H1. 
      apply inc_transitive with (2 := proj2 (subst_cst_rimage _)),
            inc_transitive with (1 := prd_image_inc _ _ subst_cst _), 
            pr_image_inc, pmo_inc.
    apply is_pmo_postfixpoint in H0.
    apply cl_inc_closed with (3 := H0) in H1; auto.
    intros (x,y) ((m,n),(Hmn,Hxy)).
    rewrite <- Hxy; apply (H1 _ Hmn).
  Qed.
  
  Proposition subst_pmo_prop m n : pmo C (m ~~ n) -> pmo (subst_dimage C) (subst_w sigma m ~~ subst_w sigma n).
  Proof. intro; apply subst_dimage_pmo_inc; exists (m,n); tauto. Qed.
  
End Substitutions.

Proposition subst_rimage_id {L} (C : _ -> Prop) m n : C (m ~~ n) -> subst_rimage (fun x : L => l x) C (m ~~ n).
Proof. intros; red; rewrite -> subst_w_id, -> subst_w_id; auto. Qed.

Proposition subst_dimage_id' {L} C : subst_dimage (fun x : L => l x) C ~ C.
Proof.
  intros. split; intros (x,y) H.
  
  destruct H as [ (x',y') (H1,H2) ].
  simpl in H2.
  rewrite subst_w_id, subst_w_id in H2. 
  rewrite <- H2;  trivial.
  
  exists (x,y). 
  split; auto. 
  simpl; rewrite subst_w_id, subst_w_id; auto.
Qed.

