(**************************************************************)
(*   The Strong Completeness of Partial Monoidal BI           *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.
Require Import Permutation.
Require Import Arith.
Require Import Peano_dec.

Require Import predicate.
Require Import enum.
Require Import enumerations.
Require Import carrier.
Require Import finite.
Require Import maps.

Reserved Notation " x 'o' y " (at level 32, left associativity).
Reserved Notation " 'l' x " (at level 31, left associativity).
Reserved Notation " x 'weq' y " (at level 50, no associativity).

Section words.

  Set Implicit Arguments.

  Variable letter : Type.

  Definition word : Type := list letter.

  Section word_eq_dec.
  
    Hypothesis letter_enum : enumeration letter.

    Definition enum_word : enumeration word.
    Proof. apply list_enum; auto. Qed.

    Proposition word_eq_dec (x y : word) : { x = y } + { x <> y }.
    Proof. apply enumeration_dec, enum_word. Qed.

  End word_eq_dec.

  Definition word_eq (x y : word) := Permutation x y.

  Delimit Scope word_scope with word.

  Notation "x 'o' y" := (x++y : word) (at level 32, left associativity) : word_scope.
  Notation eps := (nil : word).
  Notation "'l' x" := (x::nil : word) (at level 31, left associativity) : word_scope.

  (* Words are taken equivalent up to permutation of labels *)

  Infix "weq" := word_eq (at level 50, no associativity).

  Open Scope word_scope.

  Proposition eps_neutral_l x : eps o x = x.
  Proof. auto. Qed.
  
  Proposition eps_neutral_r x : x o eps = x.
  Proof. simpl; symmetry; apply app_nil_end. Qed.
  
  Proposition o_assoc x y z : (x o y) o z = x o (y o z).
  Proof. apply app_ass. Qed.
  
  Hint Immediate eps_neutral_l eps_neutral_r o_assoc.
  
  Proposition weq_rfl x : x weq x.
  Proof. apply Permutation_refl. Qed.
  
  Proposition weq_sym x y : x weq y -> y weq x.
  Proof. apply Permutation_sym. Qed.
  
  Proposition weq_tran x y z : x weq y -> y weq z -> x weq z.
  Proof. apply Permutation_trans. Qed.
  
  Proposition weq_cgr_l k x y : x weq y -> k o x weq k o y.
  Proof. apply Permutation_app_head. Qed.
  
  Proposition weq_cgr_r k x y : x weq y -> x o k weq y o k.
  Proof. apply Permutation_app_tail. Qed.
  
  Proposition weq_comm x y : x o y weq y o x.
  Proof. apply Permutation_app_swap. Qed.
  
  Hint Immediate weq_rfl weq_comm.
  Hint Resolve weq_sym weq_cgr_l weq_cgr_r.
  
  Proposition weq_cgr x x' y y' : x weq x' -> y weq y' -> x o y weq x' o y'.
  Proof. intros ? H; apply weq_tran with (2 := weq_cgr_l _ H); auto. Qed.
  
  Proposition weq_cgr_hd p x y : x weq y -> l p o x weq l p o y.
  Proof. intros; constructor; auto. Qed.
  
  Hint Resolve weq_cgr weq_cgr_hd.
  
  Proposition weq_length x y : x weq y -> length x = length y.
  Proof. apply Permutation_length. Qed.
  
  Proposition weq_carrier x y : x weq y -> carrier x ~ carrier y.
  Proof.
    intros H; 
    split; intros k Hk; [ idtac | apply weq_sym in H ];
    apply (Permutation_in k H); auto.
  Qed.

  Hint Resolve weq_carrier.
  
  Proposition carrier_eps x : carrier x inc empty -> x = eps.
  Proof.
    destruct x as [ | a x ]; auto.
    intros C.
    apply carrier_cons_1 in C.
    exfalso; tauto.
  Qed.
  
  Proposition carrier_split p x : carrier x p -> exists x', x weq l p o x'.
  Proof.
    intros H. 
    unfold carrier in H; apply in_split in H.
    destruct H as [ l1 [ l2 H ] ].
    exists (l1++l2).
    rewrite H.
    cutrewrite (p::l2 = (p::nil)++l2); auto.
    rewrite <- o_assoc, <- o_assoc.
    apply weq_cgr; auto.
  Qed.
  
  Proposition split_carrier p x x' : x weq l p o x' -> carrier x p.
  Proof.
    intros H.
    apply weq_carrier in H.
    apply H.
    simpl; auto.
  Qed.
  
  Proposition weq_split_letter a x y : l a weq x o y -> (l a weq x /\ y weq nil) 
                                                      + (l a weq y /\ x weq nil).
  Proof.
    intros H.
    destruct x as [ | x lx ].
    right; auto.
    destruct y as [ | y ly ].
    rewrite <- app_nil_end in H; left; auto.    
    apply weq_length in H.
    simpl in H; injection H; clear H; intros H.
    rewrite -> app_length in H.
    simpl in H.
    contradict H; omega.
  Qed.
  
  Proposition weq_simpl_l a x y : a o x weq a o y -> x weq y.
  Proof. intros H; apply Permutation_app_inv_l in H; auto. Qed.
  
  Proposition weq_simpl_r a x y : x o a weq y o a -> x weq y.
  Proof.
    intros H;
    apply weq_simpl_l with (a := a),
          weq_tran with (1 := weq_comm _ _),
          weq_tran with (1 := H); auto.
  Qed.

  Proposition weq_split : forall a b x y, x o y weq a o b -> exists xa, 
                                                             exists xb, 
                                                             exists ya, 
                                                             exists yb, x weq xa o xb 
                                                                     /\ y weq ya o yb 
                                                                     /\ a weq xa o ya 
                                                                     /\ b weq xb o yb.
  Proof.
    induction a as [ | p a ]; intros b x y H; simpl in H.
    exists eps; exists x; exists eps; exists y; auto.
    
    generalize (split_carrier H); intros H';
    apply carrier_app, bin_union_sg in H';
    destruct H' as [ H' | H' ].
    
    destruct (carrier_split _ _ H') as [ x' Hx' ].
    assert (x' o y weq a o b) as H''.
      apply (weq_simpl_l (l p)).
      apply weq_tran with (2 := H).
      rewrite -> ass_app; auto.
    apply IHa in H''.
    destruct H'' as ( xa' & xb & ya & yb & H1 & H2 & H3 & H4 ).
    exists ((l p) o xa'); exists xb; exists ya; exists yb.
    split. 
    apply weq_tran with (1 := Hx'); rewrite <- ass_app; auto.
    split; auto.
    split; auto.
    rewrite <- ass_app.
    cutrewrite (p::a = l p o a); auto.
    
    destruct (carrier_split _ _ H') as [ y' Hy' ].
    assert (x o y' weq a o b) as H''.
      apply weq_simpl_l with (a := l p),
            weq_tran with (2 := H),
            weq_tran with (1 := weq_comm _ _).
      rewrite <- ass_app.
      apply weq_cgr_l,
            weq_sym,
            weq_tran with (1 := Hy'),
            weq_comm.
    apply IHa in H''.
    destruct H'' as ( xa & xb & ya' & yb & H1 & H2 & H3 & H4 ).
    exists xa; exists xb; exists ((l p) o ya'); exists yb.
    split; auto.
    split. 
    apply weq_tran with (1 := Hy'). 
    rewrite <- ass_app.
    apply weq_cgr_l; auto.
    split; auto.
    apply weq_tran with (2 := weq_comm _ _).
    rewrite <- ass_app.
    cutrewrite (p::a = l p o a); auto.
    apply weq_cgr_l, weq_tran with (2 := weq_comm _ _); auto.
  Qed.
  
  Definition subword (x y : word) := exists k, x o k weq y.
  
  Proposition subword_carrier x y : subword  x y -> carrier x inc carrier y.
  Proof.
    intros [k H].
    apply weq_carrier in H.
    apply inc_transitive with (2 := proj1 H).
    auto.
  Qed.
  
  Proposition subword_split x a b : subword x (a o b) -> exists xa, exists xb, x weq xa o xb /\ subword xa a /\ subword xb b.
  Proof.
    intros [ y H ].
    apply weq_split in H.
    destruct H as ( xa & xb & ya & yb & H1 & H2 & H3 & H4 ).
    exists xa; exists xb.
    split; auto.
    split; [ exists ya | exists yb ]; auto. 
  Qed.
  
  Proposition carrier_inc_subletter m a : carrier m a -> subword (l a) m.
  Proof.
    intros H.
    apply carrier_split in H.
    destruct H as [ x H ].
    exists x; auto.
  Qed.

  Proposition subletter_inc_carrier m a : subword (l a) m -> (carrier m) a.
  Proof.
    intros[ k Hk ].
    apply weq_carrier in Hk.
    apply Hk; simpl; auto.
  Qed.

End words.

Delimit Scope word_scope with word.

Notation "x 'o' y" := (x++y : word _) (at level 32, left associativity) : word_scope.
Notation eps := (nil : word _).
Notation "'l' x" := (x::nil : word _) (at level 31, left associativity) : word_scope.
Infix "weq" := word_eq (at level 50, no associativity) : word_scope.

Open Scope word_scope.

Hint Immediate eps_neutral_l eps_neutral_r o_assoc.
Hint Immediate weq_rfl weq_comm.
Hint Resolve weq_sym weq_cgr_l weq_cgr_r.
Hint Resolve weq_cgr weq_cgr_hd.
Hint Resolve weq_carrier.

Section Substitutions.

  Set Implicit Arguments.
 
  Variable L M : Type.

  Variable sigma : L -> word M.

  Fixpoint subst_w m := match m with
    | nil => nil
    | x::m => (sigma x) o (subst_w m)
  end.

  Definition subst_alphabet (P : L -> Prop) a := exists m, carrier m inc P /\ carrier (subst_w m) a. 

  Proposition subst_o m m' : subst_w (m o m') = (subst_w m) o (subst_w m').
  Proof.
    induction m; auto. 
    simpl; rewrite -> o_assoc, <- IHm; auto.
  Qed.

  Proposition subst_carrier m x a : carrier m x -> carrier (sigma x) a -> carrier (subst_w m) a.
  Proof.
    intros H1.
    generalize a; clear a.
    induction m as [ | b m ]; intros a H2.
    contradiction H1.
    simpl in H1.
    simpl.
    apply carrier_app. 
    destruct H1 as [ H1 | H1 ].
    rewrite -> H1; apply bin_union_inc_r1; auto.
    apply bin_union_inc_r2, IHm; auto.
  Qed.

End Substitutions.

Section Substitutions_prop_1.

  Set Implicit Arguments.

  Variable L M : Type.

  Proposition subst_w_id (m : word L) : subst_w (fun x => l x) m = m.
  Proof.
    induction m; auto.
    simpl; rewrite -> IHm; auto.
  Qed.
  
  Variable s : L -> M.
  
  Proposition subst_w_carrier m a : carrier (subst_w (fun x => l (s x)) m) a -> exists b, carrier m b /\ s b = a.
  Proof.
    intros H; induction m as [ | x m ]; auto.
    contradiction H.
    simpl in H.
    destruct H as [ H | H ].
    exists x; auto.
    apply IHm in H.
    destruct H as ( b & H1 & H2 ); exists b; auto.
  Qed.

  Definition subst_alphabet_one_id P : subst_alphabet (fun x => l (s x)) P ~ pd_image s P.
  Proof.
    split.
    intros a ( m & H1 & H2 ).
    apply subst_w_carrier in H2.
    destruct H2 as (b & H2 & H3).
    exists b; auto.
    intros x (y & H1 & H2).
    rewrite <- H2; clear x H2.
    exists (y::nil).
    split; auto.
    simpl; auto.
  Qed.

End Substitutions_prop_1.

Section Subst_alphabet.

  Set Implicit Arguments.

  Variable L M : Type.

  Variable P : L -> Prop.
  Variable s s' : L -> word M.
  Hypothesis s_eq_s' : forall a, P a -> s a = s' a.

  Proposition subst_carrier_eq m : carrier m inc P -> subst_w s m = subst_w s' m.
  Proof.
    induction m; simpl; intros; auto.
    rewrite <- s_eq_s', <- IHm; auto.
    apply inc_transitive with (2 := H).
    intro; auto.
  Qed.

  Proposition subst_alphabet_prop a : subst_alphabet s P a -> exists b, P b /\ carrier (s b) a.
  Proof.
    intros [ m Hm ].
    induction m.
    destruct Hm as [ _ Hm ]. simpl in Hm; contradiction Hm.
    destruct Hm as [ Hm Hm' ]; simpl in Hm'.
    apply carrier_app in Hm'.
    apply bin_union_sg in Hm'.
    destruct Hm' as [ Hm' | Hm' ].
    exists a0. split; auto. 
    destruct IHm as [ b Hb ]. split; auto. 
    apply inc_transitive with (2:= Hm).
    intro; auto.
    exists b; auto.
  Qed.

End Subst_alphabet.

Proposition subst_alphabet_cup { L M P Q } {s : L -> word M} : subst_alphabet s (P cup Q) ~ subst_alphabet s P cup subst_alphabet s Q.
Proof.
  split.
  intros a Ha. apply subst_alphabet_prop in Ha. destruct Ha as [ b [ Hb Ha ] ].
  apply bin_union_sg in Hb. 
  destruct Hb as [ Hb | Hb ]; [ apply bin_union_inc_r1 | apply bin_union_inc_r2 ];
  exists (l b); simpl; split; auto; rewrite app_nil_r; auto.
  apply bin_union_inc_l; intros a [ m [ H1 H2 ] ]; exists m; 
  split; auto; apply inc_transitive with (1 := H1); auto.
Qed.

Section Subst_weq.

  Set Implicit Arguments.

  Variable L M : Type.

  Variable s s' : L -> word M.
  Hypothesis s_eq_s' : forall a, s a weq s' a.

  Proposition subst_weq m : subst_w s m weq subst_w s' m.
  Proof.
    induction m.
    simpl. apply weq_rfl.
    simpl. 
    apply weq_tran with (y := s a o subst_w s' m); auto.
  Qed.

End Subst_weq.




    
    